from matplotlib.pyplot import axis
import yaml
import numpy as np
import torch
import torch.nn as nn


class EPNet(nn.Module):
    def __init__(self, config):
        super(EPNet, self).__init__()
        self.config = config
        self.cmap_dim = config['cmap_sector_num']
        self.act_dim = config['act_dim']
        self.history_length = config['epnet_history_length']

        self.layer1 = nn.Sequential(
            nn.Linear(self.act_dim * self.history_length, 256),
            nn.ReLU(),
        )
        self.layer2 = nn.Sequential(
            nn.Linear(self.cmap_dim, 512),
            nn.ReLU(),
        )
        self.layer3 = nn.Sequential(
            nn.Linear(self.act_dim, 256),
            nn.ReLU(),
        )
        self.layer4 = nn.Sequential(
            nn.Linear(256 + 512 + 256, 512),
            nn.ReLU(),
        )
        self.layer5 = nn.Sequential(
            nn.Linear(512, 512),
            nn.ReLU(),
        )
        self.layer6 = nn.Sequential(
            nn.Linear(512, self.cmap_dim),
        )

    def forward(self, vc, a, vel):
        act = self.layer1(a)
        vc = self.layer2(vc)
        vel = self.layer3(vel)
        x = torch.cat((vc, act, vel), dim=-1)
        x = self.layer4(x)
        x = self.layer5(x)
        x = self.layer6(x)

        return x

    def predict(self, vc, a, vel):
        x = self.forward(vc, a, vel)
        val = self.calc_sensing_val(x)
        val /= self.config['cmap_sector_num'] # Normalize: sum to 1
        val = val.unsqueeze(-1)

        return val

    def calc_sensing_val(self, vc):
        res = vc.sum(axis=-1)

        return res

if __name__ == '__main__':
    config = None
    with open('configs.yaml') as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
    epnet = EPNet(config)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    batch_size = 64
    seq_len = 16
    vc_data = torch.randn(batch_size, seq_len, 60).to(device)
    action_data = torch.randn(batch_size, seq_len, 2).to(device)
    vel_data = torch.randn(batch_size, seq_len, 2).to(device)
    epnet.to(device)

    pred_val = epnet.predict(vc_data, action_data, vel_data)
    print(pred_val.size())