import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions import Normal


###########################################################################################################
#                                               Below Networks for SAC
###########################################################################################################

def generate_disturb_mask(obs):
    disturb_mask = torch.ones_like(obs)
    disturb_mask[obs == 0] = 0
    return disturb_mask

class QNet(nn.Module):
    def __init__(self, config):
        super(QNet, self).__init__()
        self.sensor_dim = config['obs_dim']
        self.state_dim = config['state_dim']
        self.act_dim = config['act_dim']
        self.stack_N = config['stack_N']
        self.sensor_feature_dim = 512
        self.config = config
        self.activation = F.leaky_relu

        if not self.config['use_cmap']:
            self.config['cmap_hidden_size'] = 0
        else:
            self.cmap_layer = nn.Linear(self.config['cmap_sector_num'], self.config['cmap_hidden_size'])

        self.sensor_layer = nn.Sequential(
            nn.Conv1d(2, 32, 1),
            nn.LeakyReLU(),
            nn.Conv1d(32, 64, 1),
            nn.LeakyReLU(),
            nn.Conv1d(64, 512, 1),
            nn.LeakyReLU(),
        )
        self.state_layer = nn.Sequential(
            nn.Linear(4, 32),
            nn.LeakyReLU(),
            nn.Linear(32, 64),
            nn.LeakyReLU(),
        )
        self.action_layer = nn.Sequential(
            nn.Linear(2, 32),
            nn.LeakyReLU(),
            nn.Linear(32, 64),
            nn.LeakyReLU(),
        )
        self.concat_layer = nn.Sequential(
            nn.Linear(512 + 64 + 64 + self.config['cmap_hidden_size'], 256),
            nn.LeakyReLU(),
            nn.Linear(256, 1)
        )
        
    def forward(self, obs, action):
        if self.config['use_cmap']:
            sensor, state, vc = obs[:, :self.sensor_dim * 2], \
                                obs[:, self.sensor_dim * 2 : self.sensor_dim * 2 + self.state_dim], \
                                obs[:, self.sensor_dim * 2 + self.state_dim: ]
        else:
            sensor, state = obs[:, :self.sensor_dim * 2], \
                            obs[:, self.sensor_dim * 2 : self.sensor_dim * 2 + self.state_dim]

        sensor = sensor.view(-1, self.sensor_dim, 2).transpose(1,2)
        if self.config['is_disturbance']:
            pts = int((1-self.config['disturb_constant_range']) * self.sensor_dim)
        else:
            pts = self.sensor_dim
        sensor = sensor[:, :, :pts]
        sensor = self.sensor_layer(sensor)
        sensor = torch.max(sensor, 2, keepdim=True)[0]
        sensor = sensor.view(-1, 512)

        state = self.state_layer(state)
        action = self.action_layer(action)

        if self.config['use_cmap']:
            vc = self.cmap_layer(vc)
            q_out = torch.cat([sensor, state, action, vc], dim=-1)
        else:
            q_out = torch.cat([sensor, state, action], dim=-1)
        
        q_out = self.concat_layer(q_out)

        return q_out

class DoubleQNet(nn.Module):
    def __init__(self, config):
        super(DoubleQNet, self).__init__()
        self.q1 = QNet(config)
        self.q2 = QNet(config)

    def forward(self, obs, action):
        q1 = self.q1(obs, action)
        q2 = self.q2(obs, action)

        return q1, q2

LOG_SIG_MAX = 2
LOG_SIG_MIN = -20
epsilon = 1e-6
class PolicyNetGaussian(nn.Module):
    def __init__(self, config):
        super(PolicyNetGaussian, self).__init__()
        self.sensor_dim = config['obs_dim']
        self.state_dim = config['state_dim']
        self.act_dim = config['act_dim']
        self.stack_N = config['stack_N']
        self.sensor_feature_dim = 512
        self.config = config
        self.activation = F.leaky_relu

        if not self.config['use_cmap']:
            self.config['cmap_hidden_size'] = 0
        else:
            self.cmap_layer = nn.Linear(self.config['cmap_sector_num'], self.config['cmap_hidden_size'])

        self.sensor_layer = nn.Sequential(
            nn.Conv1d(2, 32, 1),
            nn.LeakyReLU(),
            nn.Conv1d(32, 64, 1),
            nn.LeakyReLU(),
            nn.Conv1d(64, 512, 1),
            nn.LeakyReLU(),
        )
        self.state_layer = nn.Sequential(
            nn.Linear(4, 32),
            nn.LeakyReLU(),
            nn.Linear(32, 64),
            nn.LeakyReLU(),
        )
        
        self.concat_layer = nn.Linear(512 + 64 + self.config['cmap_hidden_size'], 256)
        self.layer_mean = nn.Linear(256, 2)
        self.layer_logstd = nn.Linear(256, 2)


    def forward(self, obs):
        if self.config['use_cmap']:
            sensor, state, vc = obs[:, :self.sensor_dim * 2], \
                                obs[:, self.sensor_dim * 2 : self.sensor_dim * 2 + self.state_dim], \
                                obs[:, self.sensor_dim * 2 + self.state_dim : ]
        else:
            sensor, state = obs[:, :self.sensor_dim * 2], \
                            obs[:, self.sensor_dim * 2 : self.sensor_dim * 2 + self.state_dim]

        sensor = sensor.view(-1, self.sensor_dim, 2).transpose(1,2)
        if self.config['is_disturbance']:
            pts = int((1-self.config['disturb_constant_range']) * self.sensor_dim)
        else:
            pts = self.sensor_dim
        sensor = sensor[:, :, :pts]

        sensor = self.sensor_layer(sensor)
        sensor = torch.max(sensor, 2, keepdim=True)[0]
        sensor = sensor.view(-1, 512)

        state = self.state_layer(state)

        if self.config['use_cmap']:
            vc = self.activation(self.cmap_layer(vc))
            out = torch.cat([sensor, state, vc], dim=-1)
        else:
            out = torch.cat([sensor, state], dim=-1)
        out = self.concat_layer(out)
        a_mean = self.layer_mean(out)
        a_logstd = self.layer_logstd(out)
        a_logstd = torch.clamp(a_logstd, min=LOG_SIG_MIN, max=LOG_SIG_MAX)

        return a_mean, a_logstd
    
    def sample(self, obs, cmap=None, dummy_zero=None):
        a_mean, a_logstd = self.forward(obs)
        a_std = a_logstd.exp()

        if self.config['use_confidence_mapping']:
            if cmap is not None:
                conf_val_var = torch.bmm(cmap.unsqueeze(1), 2-cmap.unsqueeze(2)).squeeze() / (2 * self.config['cmap_sector_num'])
                conf_val_var = (self.config['conf_min'] + (self.config['conf_max'] - self.config['conf_min']) * conf_val_var)
                conf_val_var = self.config['confidence_mapping_weight'] * conf_val_var.unsqueeze(-1)
                if dummy_zero is None:
                    dummy_zero = torch.zeros_like(conf_val_var)
                else:
                    dummy_zero = dummy_zero.unsqueeze(1)
                conf_val_var_linear = torch.cat((conf_val_var, dummy_zero), axis=-1) 
                a_std = a_std + conf_val_var_linear
                conf_val_var_angular = torch.cat((dummy_zero, conf_val_var), axis=-1)
                a_std = a_std - conf_val_var_angular

                a_std = torch.clamp(a_std, min=0.000001)

        normal = Normal(a_mean, a_std)
        x_t = normal.rsample()
        action = torch.tanh(x_t)
        log_prob = normal.log_prob(x_t)

        # Enforcing action Bound
        log_prob -= torch.log(1 - action.pow(2) + epsilon)
        log_prob = log_prob.sum(1, keepdim=True)
        return action, log_prob, torch.tanh(a_mean)


if __name__ == '__main__':
    # Set configs
    import yaml
    import time
    config = None
    with open('configs.yaml') as f:
        config = yaml.load(f, Loader=yaml.FullLoader)

    # Set random data
    actor = PolicyNetGaussian(config)
    critic = QNet(config)
    batch_size = 64
    input_channel = 1
    sensor_data = torch.randn(batch_size, input_channel * 2 * config['obs_dim'])
    state_data = torch.randn(batch_size, config['state_dim'])
    vc_data = torch.randn((batch_size, 60))
    obs = torch.cat((sensor_data, state_data, vc_data), 1)

    start_time = time.time()
    action, logpi, _ = actor.sample(obs)
    q_val = critic(obs, action)
    end_time = time.time()

    print(q_val.size())
    print(f'Elapsed time: {end_time - start_time} sec')