import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions import Normal


###########################################################################################################
#                                               Below Networks for SAC
###########################################################################################################

def generate_disturb_mask(obs):
    disturb_mask = torch.ones_like(obs)
    disturb_mask[obs == 0] = 0
    return disturb_mask

class QNet(nn.Module):
    def __init__(self, config):
        super(QNet, self).__init__()
        self.sensor_dim = config['obs_dim']
        self.state_dim = config['state_dim']
        self.act_dim = config['act_dim']
        self.stack_N = config['stack_N']
        self.sensor_feature_dim = 512
        self.config = config
        self.activation = F.leaky_relu
        self.recurrent_length = self.config['recurrent_length']

        if not self.config['use_cmap']:
            self.config['cmap_hidden_size'] = 0
        else:
            self.cmap_layer = nn.Sequential(
                nn.Linear(self.config['cmap_sector_num'], self.config['cmap_hidden_size']),
                nn.LeakyReLU(),
            )

        self.sensor_layer = nn.Sequential(
            nn.Conv1d(2, 32, 1),
            nn.LeakyReLU(),
            nn.Conv1d(32, 64, 1),
            nn.LeakyReLU(),
            nn.Conv1d(64, 512, 1),
            nn.LeakyReLU(),
        )
        self.state_layer = nn.Sequential(
            nn.Linear(4, 32),
            nn.LeakyReLU(),
            nn.Linear(32, 64),
            nn.LeakyReLU(),
        )
        self.action_layer = nn.Sequential(
            nn.Linear(2, 32),
            nn.LeakyReLU(),
            nn.Linear(32, 64),
            nn.LeakyReLU(),
        )
        self.concat_layer = nn.Sequential(
            nn.Linear(512 + 64 + 64 + self.config['cmap_hidden_size'], 256),
            nn.LeakyReLU(),
        )
        self.gru_layer = nn.GRU(256, 256, batch_first=True)
        self.out_layer = nn.Sequential(
            nn.Linear(256, 256),
            nn.LeakyReLU(),
            nn.Linear(256, 1)
        )

    def forward(self, obs, action, hidden):
        if self.config['use_cmap']:
            sensor, state, vc = obs[:, :, :self.sensor_dim * 2], \
                                obs[:, :, self.sensor_dim * 2 : self.sensor_dim * 2 + self.state_dim], \
                                obs[:, :, self.sensor_dim * 2 + self.state_dim: ]
        else:
            sensor, state = obs[:, :, :self.sensor_dim * 2], \
                            obs[:, :, self.sensor_dim * 2 : self.sensor_dim * 2 + self.state_dim]

        sensor = sensor.view(-1, self.sensor_dim, 2).transpose(1,2)
        if self.config['is_disturbance']:
            pts = int((1-self.config['disturb_constant_range']) * self.sensor_dim)
        else:
            pts = self.sensor_dim
        if pts != 0:
            sensor = sensor[:, :, :pts]
        sensor = self.sensor_layer(sensor)
        sensor = torch.max(sensor, 2, keepdim=True)[0]
        sensor = sensor.view(-1, state.size(1), 512)

        state = self.state_layer(state)
        action = self.action_layer(action)

        if self.config['use_cmap']:
            vc = self.cmap_layer(vc)
            q_out = torch.cat([sensor, state, action, vc], dim=-1)
        else:
            q_out = torch.cat([sensor, state, action], dim=-1)
        q_out = self.concat_layer(q_out)
        q_out, hidden = self.gru_layer(q_out, hidden)
        q_out = self.out_layer(q_out)

        return q_out, hidden

class DoubleQNet(nn.Module):
    def __init__(self, config):
        super(DoubleQNet, self).__init__()
        self.q1 = QNet(config)
        self.q2 = QNet(config)

    def forward(self, obs, action, hidden):
        q1, hidden1 = self.q1(obs, action, hidden)
        q2, hidden2 = self.q2(obs, action, hidden)

        return q1, q2, hidden1, hidden2

LOG_SIG_MAX = 2
LOG_SIG_MIN = -20
epsilon = 1e-6
class PolicyNetGaussian(nn.Module):
    def __init__(self, config):
        super(PolicyNetGaussian, self).__init__()
        self.sensor_dim = config['obs_dim']
        self.state_dim = config['state_dim']
        self.act_dim = config['act_dim']
        self.stack_N = config['stack_N']
        self.sensor_feature_dim = 512
        self.config = config
        self.activation = F.leaky_relu
        self.recurrent_length = self.config['recurrent_length']

        if not self.config['use_cmap']:
            self.config['cmap_hidden_size'] = 0
        else:
            self.cmap_layer = nn.Sequential(
                nn.Linear(self.config['cmap_sector_num'], self.config['cmap_hidden_size']),
                nn.LeakyReLU(),
            )
        self.sensor_layer = nn.Sequential(
            nn.Conv1d(2, 32, 1),
            nn.LeakyReLU(),
            nn.Conv1d(32, 64, 1),
            nn.LeakyReLU(),
            nn.Conv1d(64, 512, 1),
            nn.LeakyReLU(),
        )
        self.state_layer = nn.Sequential(
            nn.Linear(4, 32),
            nn.LeakyReLU(),
            nn.Linear(32, 64),
            nn.LeakyReLU(),
        )
        self.concat_layer = nn.Sequential(
            nn.Linear(512 + 64 + self.config['cmap_hidden_size'], 256),
            nn.LeakyReLU(),
        )
        self.gru_layer = nn.GRU(256, 256, batch_first=True)
        self.out_layer = nn.Sequential(
            nn.Linear(256, 256),
            nn.LeakyReLU()
        )
        self.layer_mean = nn.Linear(256, 2)
        self.layer_logstd = nn.Linear(256, 2)

    def forward(self, obs, hidden):
        if self.config['use_cmap']:
            sensor, state, vc = obs[:, :, :self.sensor_dim * 2], \
                                obs[:, :, self.sensor_dim * 2 : self.sensor_dim * 2 + self.state_dim], \
                                obs[:, :, self.sensor_dim * 2 + self.state_dim : ]
        else:
            sensor, state = obs[:, :, :self.sensor_dim * 2], \
                            obs[:, :, self.sensor_dim * 2 : self.sensor_dim * 2 + self.state_dim]

        sensor = sensor.view(-1, self.sensor_dim, 2).transpose(1,2)
        if self.config['is_disturbance']:
            pts = int((1-self.config['disturb_constant_range']) * self.sensor_dim)
        else:
            pts = self.sensor_dim

        if pts != 0:
            sensor = sensor[:, :, :pts]

        sensor = self.sensor_layer(sensor)
        sensor = torch.max(sensor, 2, keepdim=True)[0]
        sensor = sensor.view(-1, state.size(1), 512)

        state = self.state_layer(state)

        if self.config['use_cmap']:
            vc = self.cmap_layer(vc)
            out = torch.cat([sensor, state, vc], dim=-1)
        else:
            out = torch.cat([sensor, state], dim=-1)
        out = self.concat_layer(out)
        out, hidden = self.gru_layer(out, hidden)
        out = self.out_layer(out)
        a_mean = self.layer_mean(out)
        a_logstd = self.layer_logstd(out)
        a_logstd = torch.clamp(a_logstd, min=LOG_SIG_MIN, max=LOG_SIG_MAX)

        return a_mean, a_logstd, hidden
    
    def sample(self, obs, hidden, cmap=None, dummy_zero=None):
        a_mean, a_logstd, hidden = self.forward(obs, hidden)
        a_std = a_logstd.exp()

        if self.config['use_confidence_mapping']:
            if cmap is not None:
                conf_val_var = torch.bmm(cmap, 2 - cmap.transpose(1,2))
                conf_val_var = conf_val_var.view(conf_val_var.size(0), -1)
                conf_val_var = conf_val_var / (2 * self.config['cmap_sector_num'])
                conf_val_var = (self.config['conf_min'] + (self.config['conf_max'] - self.config['conf_min']) * conf_val_var)
                conf_val_var = self.config['confidence_mapping_weight'] * conf_val_var.unsqueeze(-1)
                if dummy_zero is None:
                    dummy_zero = torch.zeros_like(conf_val_var)
                else:
                    dummy_zero = dummy_zero.unsqueeze(1)
                conf_val_var_linear = torch.cat((conf_val_var, dummy_zero), axis=-1) 
                # conf_val_var_linear = torch.cat((conf_val_var, dummy_zero), axis=-1) 
                a_mean = a_mean - conf_val_var_linear
                # conf_val_var_angular = torch.cat((dummy_zero, conf_val_var), axis=-1)
                # a_std = a_std - conf_val_var_angular

                a_std = torch.clamp(a_std, min=0.000001)

        normal = Normal(a_mean, a_std)
        x_t = normal.rsample()
        action = torch.tanh(x_t)
        log_prob = normal.log_prob(x_t)

        # Enforcing action Bound
        log_prob -= torch.log(1 - action.pow(2) + epsilon)
        log_prob = log_prob.sum(-1, keepdim=True)
        return action, log_prob, torch.tanh(a_mean), hidden.cpu().detach()


if __name__ == '__main__':
    # Set configs
    import yaml
    import time
    config = None
    with open('configs.yaml') as f:
        config = yaml.load(f, Loader=yaml.FullLoader)

    # Set random data
    actor = PolicyNetGaussian(config)
    critic = QNet(config)
    batch_size = 1
    input_channel = 1
    recurrent_length = 16

    # Data: (batch, seq_len, feature_dim)
    sensor_data = torch.randn(batch_size, recurrent_length, config['obs_dim']-24)
    disturb_data = torch.zeros(batch_size, recurrent_length, config['obs_dim']+24)
    sensor_data = torch.cat((sensor_data, disturb_data), dim=-1)

    state_data = torch.randn(batch_size, recurrent_length, config['state_dim'])
    vc_data = torch.randn((batch_size, recurrent_length, 60))
    obs = torch.cat((sensor_data, state_data, vc_data), -1)
    hidden = torch.randn(1, batch_size, config['recurrent_hidden_dim'])  
    # hidden_dim: [num_layers * num_directions, batch, hidden_size]

    start_time = time.time()
    action, logpi, _, hidden_out = actor.sample(obs, hidden)
    q_val, hidden_out = critic(obs, action, hidden)
    end_time = time.time()

    print(q_val.size())
    print(f'Elapsed time: {end_time - start_time} sec')