import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions import Normal
from gridmap.gridmap_autoencoder import GridEncoder

class PolicyNet(nn.Module):
    def __init__(self, config):
        super(PolicyNet, self).__init__()
        self.sensor_dim = config['obs_dim']
        self.stack_N = config['stack_N']
        self.state_dim = config['state_dim']
        self.act_dim = config['act_dim']

        self.layer1 = nn.Sequential(
            nn.Conv1d(1, 32, kernel_size=5, stride=2),
            nn.ReLU(),
        )
        self.layer2 = nn.Sequential(
            nn.Conv1d(32, 32, kernel_size=3, stride=2),
            nn.ReLU(),
        )
        
        if self.config['stack_N'] == 1:
            self.layer3 = nn.Sequential( nn.Linear(896, self.sensor_feature_dim), nn.ReLU() )
        elif self.config['stack_N'] == 3:
            self.layer3 = nn.Sequential( nn.Linear(2816, self.sensor_feature_dim), nn.ReLU() )
        elif self.config['stack_N'] == 6:
            self.layer3 = nn.Sequential( nn.Linear(5696, self.sensor_feature_dim), nn.ReLU() )
        elif self.config['stack_N'] == 10:
            self.layer3 = nn.Sequential( nn.Linear(9536, self.sensor_feature_dim), nn.ReLU() )
        else:
            raise Exception("error! Check your stack N size!")

        self.layer4 = nn.Sequential(
            nn.Linear(256 + self.state_dim, 128),
            nn.ReLU(),
        )
        self.layer5 = nn.Sequential(
            nn.Linear(128, 2)
        )

    def forward(self, obs):
        # Seperate obs into 'sensor' data and 'state' data
        obs = obs.unsqueeze(1)
        sensor, state = obs[:, :, :self.sensor_dim*self.stack_N], obs[:, :, self.sensor_dim*self.stack_N:].squeeze(1) 
        sensor = sensor.reshape(-1, 1, self.stack_N * self.sensor_dim)

        # Feed sensor data
        sensor = self.layer1(sensor)
        sensor = self.layer2(sensor)
        sensor = sensor.view(sensor.size(0), -1)  # Flatten (Batch_size * FC)
        sensor = self.layer3(sensor)

        # Concat 'sensor' data and 'state' data
        out = torch.cat([sensor, state], dim=-1)
        out = self.layer4(out)
        out = self.layer5(out)
        out = torch.tanh(out)

        return out

###########################################################################################################
#                                               Below for SAC
###########################################################################################################

class QNet(nn.Module):
    def __init__(self, config):
        super(QNet, self).__init__()
        self.sensor_dim = config['obs_dim']
        self.stack_N = config['stack_N']
        self.state_dim = config['state_dim']
        self.act_dim = config['act_dim']
        self.sensor_feature_dim = 512
        self.config = config

        if self.config['use_gridmap_encoder']:
            self.gridmap_encoder = GridEncoder(self.config)
            # device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            device = 'cpu'
            self.gridmap_encoder.load_state_dict(torch.load(self.config['gridmap_encoder_path'], map_location=device)['model'])
            self.gridmap_encoder.to(device)
            for param in self.gridmap_encoder.parameters():
                param.requires_grad = False
        else:
            self.config['gridmap_encoder_output_size'] = 0

        if not self.config['use_cmap']:
            self.config['cmap_hidden_size'] = 0

        self.layer1 = nn.Sequential(
            nn.Conv1d(1, 32, kernel_size=5, stride=2),
            nn.ReLU(),
        )
        self.layer2 = nn.Sequential(
            nn.Conv1d(32, 32, kernel_size=3, stride=2),
            nn.ReLU(),
        )

        if self.config['stack_N'] == 1:
            self.layer3 = nn.Sequential( nn.Linear(896, self.sensor_feature_dim), nn.ReLU() )
        elif self.config['stack_N'] == 3:
            self.layer3 = nn.Sequential( nn.Linear(2816, self.sensor_feature_dim), nn.ReLU() )
        elif self.config['stack_N'] == 6:
            self.layer3 = nn.Sequential( nn.Linear(5696, self.sensor_feature_dim), nn.ReLU() )
        elif self.config['stack_N'] == 10:
            self.layer3 = nn.Sequential( nn.Linear(9536, self.sensor_feature_dim), nn.ReLU() )
        else:
            raise Exception("error! Check your stack N size!")

        self.layer4 = nn.Sequential(
            nn.Linear(self.sensor_feature_dim + self.state_dim + self.act_dim + self.config['cmap_hidden_size']\
                 + self.config['gridmap_encoder_output_size'], 128),
            nn.ReLU(),
        )
        self.layer5 = nn.Sequential(
            nn.Linear(128, 1)
        )

        if self.config['use_cmap']:
            self.cmap_layer = nn.Sequential(
                    nn.Linear(self.config['cmap_sector_num'], self.config['cmap_hidden_size']),
                    nn.ReLU()
            )

    def forward(self, obs, a):
        obs = obs.unsqueeze(1)

        if self.config['use_cmap']:
            sensor, state, vc = obs[:, :, :self.sensor_dim*self.stack_N], \
                                obs[:, :, self.sensor_dim*self.stack_N:self.sensor_dim*self.stack_N+self.state_dim].squeeze(1), \
                                obs[:, :, -self.config['cmap_sector_num']:].squeeze(1)
        else:
            sensor, state = obs[:, :, :self.sensor_dim*self.stack_N], \
                            obs[:, :, self.sensor_dim*self.stack_N:self.sensor_dim*self.stack_N+self.state_dim].squeeze(1)

        sensor = sensor.reshape(-1, 1, self.stack_N * self.sensor_dim)

        # Feed sensor data
        sensor = self.layer1(sensor)
        sensor = self.layer2(sensor)
        sensor = sensor.view(sensor.size(0), -1)  # Flatten (Batch_size * FC)
        sensor = self.layer3(sensor)

        s_a = torch.cat((state, a), dim=-1)
        if self.config['use_cmap']:
            vc = self.cmap_layer(vc)
            q_out = torch.cat([sensor, s_a, vc], dim=-1)
        else:
            q_out = torch.cat([sensor, s_a], dim=-1)
        q_out = self.layer4(q_out)
        q_out = self.layer5(q_out)

        return q_out

class DoubleQNet(nn.Module):
    def __init__(self, config):
        super(DoubleQNet, self).__init__()
        self.q1 = QNet(config)
        self.q2 = QNet(config)

    def forward(self, obs, action):
        q1 = self.q1(obs, action)
        q2 = self.q2(obs, action)

        return q1, q2

LOG_SIG_MAX = 2
LOG_SIG_MIN = -20
epsilon = 1e-6
class PolicyNetGaussian(nn.Module):
    def __init__(self, config):
        super(PolicyNetGaussian, self).__init__()
        self.sensor_dim = config['obs_dim']
        self.stack_N = config['stack_N']
        self.state_dim = config['state_dim']
        self.act_dim = config['act_dim']
        self.sensor_feature_dim = 512
        self.config = config

        if self.config['use_gridmap_encoder']:
            self.gridmap_encoder = GridEncoder(self.config)
            # device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            device = 'cpu'
            self.gridmap_encoder.load_state_dict(torch.load(self.config['gridmap_encoder_path'], map_location=device)['model'])
            self.gridmap_encoder.to(device)
            for param in self.gridmap_encoder.parameters():
                param.requires_grad = False
        else:
            self.config['gridmap_encoder_output_size'] = 0

        if not self.config['use_cmap']:
            self.config['cmap_hidden_size'] = 0

        self.layer1 = nn.Sequential(
            nn.Conv1d(1, 32, kernel_size=5, stride=2),
            nn.ReLU(),
        )
        self.layer2 = nn.Sequential(
            nn.Conv1d(32, 32, kernel_size=3, stride=2),
            nn.ReLU(),
        )

        if self.config['stack_N'] == 1:
            self.layer3 = nn.Sequential( nn.Linear(896, self.sensor_feature_dim), nn.ReLU() )
        elif self.config['stack_N'] == 3:
            self.layer3 = nn.Sequential( nn.Linear(2816, self.sensor_feature_dim), nn.ReLU() )
        elif self.config['stack_N'] == 6:
            self.layer3 = nn.Sequential( nn.Linear(5696, self.sensor_feature_dim), nn.ReLU() )
        elif self.config['stack_N'] == 10:
            self.layer3 = nn.Sequential( nn.Linear(9536, self.sensor_feature_dim), nn.ReLU() )
        else:
            raise Exception("error! Check your stack N size!")

        self.layer4 = nn.Sequential(
            nn.Linear(self.sensor_feature_dim + self.state_dim + self.config['cmap_hidden_size']\
                 + self.config['gridmap_encoder_output_size'], 128),
            nn.ReLU(),
        )
        self.layer_mean = nn.Sequential(
            nn.Linear(128, 2)
        )
        self.layer_logstd = nn.Sequential(
            nn.Linear(128, 2)
        )

        if self.config['use_cmap']:
            self.cmap_layer = nn.Sequential(
                    nn.Linear(self.config['cmap_sector_num'], self.config['cmap_hidden_size']),
                    nn.ReLU()
            )

    def forward(self, obs):
        # Seperate obs into 'sensor' data and 'state' data
        obs = obs.unsqueeze(1)

        if self.config['use_cmap']:
            sensor, state, vc = obs[:, :, :self.sensor_dim*self.stack_N], \
                                obs[:, :, self.sensor_dim*self.stack_N:self.sensor_dim*self.stack_N+self.state_dim].squeeze(1), \
                                obs[:, :, -self.config['cmap_sector_num']:].squeeze(1)
        else:
            sensor, state = obs[:, :, :self.sensor_dim*self.stack_N], \
                            obs[:, :, self.sensor_dim*self.stack_N:self.sensor_dim*self.stack_N+self.state_dim].squeeze(1)

        sensor = sensor.reshape(-1, 1, self.stack_N * self.sensor_dim)

        # Feed sensor data
        sensor = self.layer1(sensor)
        sensor = self.layer2(sensor)
        sensor = sensor.view(sensor.size(0), -1)  # Flatten (Batch_size * FC)
        sensor = self.layer3(sensor)

        if self.config['use_cmap']:
            vc = self.cmap_layer(vc)
            out = torch.cat([sensor, state, vc], dim=-1)
        else:
            out = torch.cat([sensor, state], dim=-1)
        out = self.layer4(out)
        a_mean = self.layer_mean(out)
        a_logstd = self.layer_logstd(out)
        a_logstd = torch.clamp(a_logstd, min=LOG_SIG_MIN, max=LOG_SIG_MAX)

        return a_mean, a_logstd
    
    def sample(self, obs, cmap=None, dummy_zero=None):
        a_mean, a_logstd = self.forward(obs)
        a_std = a_logstd.exp()

        if self.config['use_confidence_mapping']:
            if cmap is not None:
                conf_val_var = torch.bmm(cmap.unsqueeze(1), 2-cmap.unsqueeze(2)).squeeze() / (2 * self.config['cmap_sector_num'])
                conf_val_var = (self.config['conf_min'] + (self.config['conf_max'] - self.config['conf_min']) * conf_val_var)
                conf_val_var = self.config['confidence_mapping_weight'] * conf_val_var.unsqueeze(-1)
                if dummy_zero is None:
                    dummy_zero = torch.zeros_like(conf_val_var)
                else:
                    dummy_zero = dummy_zero.unsqueeze(1)
                conf_val_var_linear = torch.cat((conf_val_var, dummy_zero), axis=-1) 
                a_std = a_std + conf_val_var_linear
                conf_val_var_angular = torch.cat((dummy_zero, conf_val_var), axis=-1)
                a_std = a_std - conf_val_var_angular

                a_std = torch.clamp(a_std, min=0.000001)

        normal = Normal(a_mean, a_std)
        x_t = normal.rsample()
        action = torch.tanh(x_t)
        log_prob = normal.log_prob(x_t)

        # Enforcing action Bound
        log_prob -= torch.log(1 - action.pow(2) + epsilon)
        log_prob = log_prob.sum(1, keepdim=True)
        return action, log_prob, torch.tanh(a_mean)


# def get_disturb_mask(config, obs):
#     mask = torch.ones_like(obs)
#     mask[obs < config['sensor_static_disturb_mask_length']] = 0
#     return mask

if __name__ == '__main__':
    # Set configs
    import yaml
    import time
    config = None
    with open('configs.yaml') as f:
        config = yaml.load(f, Loader=yaml.FullLoader)

    # Set random data
    if config['algo_name'].lower() == 'ddpg':
        actor = PolicyNet(config)
    elif config['algo_name'].lower() == 'sac' :
        actor = PolicyNetGaussian(config)
    critic = QNet(config)
    batch_size = 64
    input_channel = 1
    sensor_data = torch.randn(batch_size, input_channel * config['stack_N'] * config['obs_dim'])
    state_data = torch.randn(batch_size, config['state_dim'])
    # gridmap_data = torch.randn(batch_size, 60*60)
    # obs = torch.cat((sensor_data, state_data, gridmap_data), 1)
    obs = torch.cat((sensor_data, state_data), 1)

    start_time = time.time()
    if config['algo_name'].lower() == 'ddpg':
        action = actor(obs)
        q_val = critic(obs, action)
    elif config['algo_name'].lower() == 'sac' :
        action, logpi, _ = actor.sample(obs)
        q_val = critic(obs, action)
        import torch.optim as optim
        critic = DoubleQNet(config)
        critic_param = critic.parameters()
    end_time = time.time()

    print(q_val.size())
    print(f'Elapsed time: {end_time - start_time} sec')