from copy import deepcopy
import numpy as np
import cv2
import matplotlib.pyplot as plt
import math
import copy
from mobile2d.utils import *
# from statsmodels.nonparametric.smoothers_lowess import lowess

class ConfidenceMap:
    def __init__(self, img_map, config):
        self.cmap_resolution = config['cmap_resolution']
        self.m = np.zeros((int(img_map.shape[0]/self.cmap_resolution), int(img_map.shape[1]/self.cmap_resolution)))
        self.config = config
        self.center = None
        self.yaw = None
        self.cmap_radius = int(round(self.config['cmap_radius'] / self.config['cmap_resolution']))
        self.cmap_cover_size = self.cmap_radius * 2
        self.cmap_y, self.cmap_x = np.ogrid[:self.cmap_cover_size, :self.cmap_cover_size]
        self.cmap_cover = np.zeros((self.cmap_cover_size, self.cmap_cover_size))

        # for parsing cmap
        self.width_left, self.height_upper = int(self.cmap_cover.shape[1] // 2), int(self.cmap_cover.shape[0] // 2)
        self.width_right, self.height_lower = int(self.cmap_cover.shape[1] - self.width_left), int(self.cmap_cover.shape[0] - self.height_upper)

        # for extracting circular shape cmap from rectangle cmap_cover
        dist_from_center = np.sqrt((self.cmap_x - self.cmap_radius)**2 + (self.cmap_y-self.cmap_radius)**2)
        self.cmap_circle_mask = dist_from_center <= self.cmap_radius

        # for masking heading direction
        self.doi_sector_masks = self._get_doi_sector_masks()

        # for fixing ROI window
        self.timestep = 0

        # for recording init pos. of the robot
        self.is_init_step = True
        self.init_pos = None
        self.dist2goal_ratio = -1
        self.init_dist = -1

        # for evaluating sensed percent
        self.sensed_val = -1
        self.sensed_val_list = []

    def initialize(self, goal_pos=None):
        self.m = np.zeros_like(self.m)
        self.center = None
        self.yaw = None
        self.grid_map = None
        self.timestep = 0
        self.is_init_step = True
        self.dist2goal_ratio = -1
        self.init_dist = -1
        self.goal_pos = goal_pos
        self.sensed_val = -1
        self.sensed_val_list = []

    def step(self, robot_pos, end_points, orien):
        if self.is_init_step:
            self.init_pos = robot_pos
            self.init_dist = self.calc_dist2goal()
            self.is_init_step = False

        self.m *= self.config['cmap_time_weight']
        self.m[self.m < self.config['cmap_sensing_threshold']] = 0
        self.update_param(robot_pos)
        self.update_map_from_end_points(end_points)

        if self.config['use_rel_window']:
            cmap = self.get_confidence_map(orien)
        else:
            cmap = self.get_confidence_map()

        self.timestep += 1
        self.dist2goal_ratio = self.calc_dist2goal(robot_pos) / self.init_dist

        self.sensed_val = cmap.sum() / cmap.shape[0]
        self.sensed_val_list.append(self.sensed_val)

        return cmap

    def get_total_sensing_conf(self):
        return sum(self.sensed_val_list) / len(self.sensed_val_list)

    def get_dist2goal_ratio(self):
        return self.dist2goal_ratio

    def calc_dist2goal(self, robot_pos=None):
        if robot_pos is None:
            dist = math.sqrt((self.init_pos[0] - self.goal_pos[0])**2
             + (self.init_pos[1] - self.goal_pos[1])**2)
        else:
            dist = math.sqrt((robot_pos[0] - self.goal_pos[0])**2
             + (robot_pos[1] - self.goal_pos[1])**2)
        return dist

    def get_gridmap(self):
        gridmap_list = self.grid_map.flatten().tolist()
        return gridmap_list

    def get_confidence_map(self, orien=None):
        sensing_cmap = self._extract_cmap()
        sensing_cmap[~self.cmap_circle_mask] = 0

        if orien is not None:
            rotation_mat = cv2.getRotationMatrix2D((self.cmap_radius, self.cmap_radius), 270 + orien, 1)
        else:
            rotation_mat = cv2.getRotationMatrix2D((self.cmap_radius, self.cmap_radius), 270 - self.yaw, 1)
        sensing_cmap = cv2.warpAffine(sensing_cmap, rotation_mat, (self.cmap_cover_size, self.cmap_cover_size), flags=cv2.INTER_LINEAR)

        # Flip back to the current image
        self.sensing_cmap = cv2.flip(sensing_cmap, 0)

        # for binary occupancy grid map
        self.grid_map = cv2.resize(self.sensing_cmap, (60, 60), interpolation=cv2.INTER_AREA)
        self.grid_map = np.where(self.grid_map < 0.4, 0.0, 1.0)

        # print("\nunique: ", end=' ')
        # print(np.unique(self.grid_map))

        # Apply sector masks to get each value of the sector
        sensing_confidence_sectors = np.array(self._apply_sector_masks(self.sensing_cmap))

        return sensing_confidence_sectors

    def draw_roi_rel_angle(self, img, robot_center, rel_angle):
        if self.config['use_rel_window']:
            base_angle = 90 + rel_angle
        else:
            base_angle = 90 - self.yaw
        start_angle = 270 - self.config['cmap_doi_angle'] // 2
        end_angle = 270 + self.config['cmap_doi_angle'] // 2
        roi_color = (0,0,255)
        roi_thinkness = 2

        img = cv2.ellipse(img, robot_center, (self.cmap_radius, self.cmap_radius), base_angle, 
                start_angle, end_angle, roi_color, roi_thinkness)

        start_xy = self._polar2cart(self.cmap_radius, math.radians(start_angle+base_angle))
        start_xy = (start_xy[0] + robot_center[0], start_xy[1] + robot_center[1])
        end_xy = self._polar2cart(self.cmap_radius, math.radians(end_angle+base_angle))
        end_xy = (end_xy[0] + robot_center[0], end_xy[1] + robot_center[1])

        img = cv2.line(img, robot_center, start_xy, roi_color, roi_thinkness)
        img = cv2.line(img, robot_center, end_xy, roi_color, roi_thinkness)

        return img

    def calc_sensing_val(self, vc):
        return vc.sum()

    def get_conf_from_cur_pos(self, robot_pos):
        def _rot_pos(x,y,phi_):
            phi = np.deg2rad(phi_)
            return np.array((x*np.cos(phi)+y*np.sin(phi), -x*np.sin(phi)+y*np.cos(phi)))

        x, y, yaw = robot_pos
        radius = 20
        dist_from_robot = 30
        center = _rot_pos(dist_from_robot, 0, -yaw) + np.array((x,y))
        cx, cy = tuple(map(int, (map(round, [center[0], center[1]]) ) ) )

        img = copy.deepcopy(self.m)

        h, w = self.m.shape[0], self.m.shape[1]
        y, x = np.ogrid[:h, :w]
        dist_from_center = np.sqrt((x - cx)**2 + (y-cy)**2)
        mask = dist_from_center <= radius

        img[~mask] = 0
        
        conf_val = img.sum() / np.count_nonzero(mask)

        return conf_val


    def update_param(self, robot_pos):
        self.center = (int(round(robot_pos[0] / self.config['cmap_resolution'])), int(round(robot_pos[1] / self.config['cmap_resolution'])))
        self.yaw = 360-robot_pos[2]

    def update_map_from_end_points(self, end_points):
        pos = self.center
        for pts in end_points:  # length of end_points: sensor_size
            cv2.line(self.m, (int(1*pos[0]), int(1*pos[1])), (int(1*pts[0]), int(1*pts[1])), (1.0,1.0,1.0), 4, lineType=cv2.LINE_AA)
        # self.m += current_cmap
    
    def _extract_cmap(self):
        lower_pad = left_pad = right_pad = upper_pad = 0
        upper_idx = int(self.center[1]) + self.height_upper
        if upper_idx > self.m.shape[0]-1:
            upper_pad = abs(self.m.shape[0]-1 - upper_idx)
            upper_idx = self.m.shape[0]-1
        lower_idx = int(self.center[1]) - self.height_lower
        if lower_idx < 0:
            lower_pad = abs(lower_idx)
            lower_idx = 0
        left_idx = int(self.center[0]) - self.width_left
        if left_idx < 0:
            left_pad = abs(left_idx)
            left_idx = 0
        right_idx = int(self.center[0]) + self.width_right
        if right_idx > self.m.shape[1]-1:
            right_pad = abs(self.m.shape[1]-1 - right_idx)
            right_idx = self.m.shape[1]-1
        
        cmap_boundary = self.m[lower_idx:upper_idx, left_idx:right_idx]
        cmap_rect = np.zeros_like(self.cmap_cover)
        cmap_rect[lower_pad:self.cmap_cover_size-upper_pad, left_pad:self.cmap_cover_size-right_pad] = cmap_boundary

        return cmap_rect

    def _apply_sector_masks(self, cmap):
        """
        Apply sector masks to cmap 
        and return 'sector_num' lists containing summed up of each sector value.
        """
        sector_list = []
        for sector_mask in self.doi_sector_masks:
            tmp_cmap = np.array(cmap)
            tmp_cmap[~sector_mask] = 0
            conf_val = tmp_cmap.sum() / np.count_nonzero(sector_mask)
            sector_list.append(conf_val)

        for i in range(len(sector_list)):
            if sector_list[i] < 0.15:
                sector_list[i] = 0
            elif sector_list[i] > 0.85:
                sector_list[i] = 1

        return sector_list

    def _get_doi_sector_masks(self):
        masks = []
        target_angle = self.config['cmap_doi_angle']
        sector_num = self.config['cmap_sector_num']
        offset_angle = (180 - target_angle) // 2
        inter = target_angle // sector_num
        for sector in range(sector_num):
            start_angle = 180 + offset_angle + sector * inter
            end_angle = 180 + offset_angle + (sector + 1) * inter
            mask = self._sector_mask(self.cmap_cover.shape, (self.cmap_radius, self.cmap_radius), self.cmap_radius,(start_angle, end_angle))
            masks.append(mask)
        return masks

    def _sector_mask(self, shape,center,radius,angle_range):
        """
        Return a boolean mask for a circular sector. The start/stop angles in  
        `angle_range` should be given in clockwise order.
        """
        x,y = np.ogrid[:shape[0],:shape[1]]
        cx,cy = center
        tmin,tmax = np.deg2rad(angle_range)
        # ensure stop angle > start angle
        if tmax < tmin:
            tmax += 2*np.pi
        r2 = (x-cx)*(x-cx) + (y-cy)*(y-cy)
        theta = np.arctan2(x-cx,y-cy) - tmin
        # wrap angles between 0 and 2*pi
        theta %= (2*np.pi)
        # circular mask

        circmask =  (r2 >= 1000) & (r2 <= radius*radius)

        anglemask = theta <= (tmax-tmin)

        return circmask*anglemask

    def _cart2polar(self, x, y):
        rho = np.sqrt(x**2 + y**2)
        phi = np.arctan2(y, x)
        return (rho, phi)

    def _polar2cart(self, radius, theta):
        x = int(round(radius * math.cos(theta)))
        y = int(round(radius * math.sin(theta)))
        return (x, y)