from operator import sub
import numpy as np
import random
import copy


class DisturbedSensor:
    def __init__(self, config):
        self.config = config
        self.generate_prob = config['disturb_generate_prob']
        self.width_range = config['disturb_width_range']
        self.length_range = config['disturb_length_range']
        self.fluctuation = config['disturb_fluctuation']
        self.max_num = config['disturb_max_num']
        self.value_range = config['disturb_value_range']
        self.constant_range = config['disturb_constant_range']
        self.cur_ds_num = 0
        self.disturbed_sensor_set = []
        self.turn_on_const_disturb = False

        self.chosen_disturb_sensor_config = -1
        self.is_selected_random_config = False
        self.num_of_random_sensor_config = len(config['disturb_mask'])
        self.mode = self.config['disturb_mode']
        self.pcl_filter_val = config['pcl_filter_val']

        self.timestep = 0

        self.set_disturb_range()
        
    def set_disturb_range(self):
        if self.config['disturb_random_range_use']:
            min_val, max_val = self.config['disturb_random_range']
            self.constant_range = np.random.uniform(min_val, max_val)

    def trigger_disturbance(self, sensor):
        if not isinstance(sensor, np.ndarray):
            sensor = np.array(sensor)
        
        for i, ds in enumerate(self.disturbed_sensor_set):
            start_loc, end_loc, val, length, width = ds['start_loc'], ds['end_loc'], ds['val'], ds['left_len'], ds['width']
            fluctuation = random.randint(-self.fluctuation, self.fluctuation)
            if fluctuation < 0:
                if start_loc + fluctuation >= 0:
                    start_loc += fluctuation
                    end_loc += fluctuation
                else:
                    start_loc = 0
                    end_loc = width
            elif fluctuation > 0:
                if end_loc + fluctuation <= self.config['obs_dim']-1:
                    start_loc += fluctuation
                    end_loc += fluctuation
                else:
                    start_loc = self.config['obs_dim']-1-width
                    end_loc = self.config['obs_dim']-1
            sensor[start_loc:end_loc] = sensor[start_loc:end_loc]*val
            length -= 1
            if length == 0:
                del self.disturbed_sensor_set[i]
                self.cur_ds_num -= 1
            else:
                self.disturbed_sensor_set[i] = ({'start_loc':start_loc, 'end_loc':end_loc, \
                        'val':val, 'left_len': length, 'width':width})      
        
        maxnum_new_sensor = self.max_num-self.cur_ds_num
        new_sensor_num = random.randint(0, maxnum_new_sensor)

        for _ in range(new_sensor_num):
            generate = random.uniform(0,1) > 1-self.generate_prob
            if generate:
                width = random.randint(2, self.width_range)
                start_loc = random.randint(0, self.config['obs_dim']-1-width)
                val = random.uniform(0, self.value_range)
                length = random.randint(1, self.length_range)
                end_loc = start_loc + width
                sensor[start_loc:end_loc] = sensor[start_loc:end_loc]*val

                if length > 1:
                    self.disturbed_sensor_set.append({'start_loc':start_loc, 'end_loc':end_loc, \
                        'val':val, 'left_len': length-1, 'width':width})
                    self.cur_ds_num += 1

        return sensor

    def divide_num_with_probs(self, num, probs):
        divided_num_lst = [int(round(num * p)) for p in probs]
        rest = num - sum(divided_num_lst)
        if rest > 0:
            divided_num_lst[divided_num_lst.index(min(divided_num_lst))] += rest
        elif rest < 0:
            divided_num_lst[divided_num_lst.index(max(divided_num_lst))] += rest

        return divided_num_lst

    def generate_random_probs(self, multi_num):
        probs = np.ones(multi_num)
        noise = 0.2
        probs[0] += noise
        probs[-1] -= noise
        probs /= probs.sum()

        return probs

    def get_mask_idx(self, divided_num_lst):
        mask_idx = []
        obs_idx = [i for i in range(self.config['obs_dim'])]
        obs_sub_idx_list = [copy.deepcopy(obs_idx)]
        divided_num_lst.sort(reverse=True)

        for num in divided_num_lst:
            random.shuffle(obs_sub_idx_list)
            for i, sub_idx_list in enumerate(obs_sub_idx_list):
                possible_size = len(sub_idx_list) - num
                if possible_size > 0:
                    start_idx = np.random.randint(0, possible_size)
                    end_idx = start_idx + num
                    obs_sub_idx_list[i] = sub_idx_list[:start_idx]
                    obs_sub_idx_list.insert(i + 1, sub_idx_list[end_idx:])

                    mask_idx.append(sub_idx_list[0] + start_idx)
                    break

        assert len(divided_num_lst) == len(mask_idx)

        return mask_idx

    def trigger_constant_damage(self, sensor, mode='left'):
        if not isinstance(sensor, np.ndarray):
            sensor = np.array(sensor)

        occluded_size = int(self.constant_range * self.config['obs_dim'])
        masks = self.config['disturb_mask']
        mask = np.zeros_like(sensor)

        if self.config['disturb_enable_dynamic'] and not self.is_selected_random_config:
            self.is_selected_random_config = True
            self.mode = 'dynamic'
            self.dynamic_disturb_time = random.randint(12, 25)

        if not self.is_selected_random_config:
            self._update_random_sensor()
            self.is_selected_random_config = True
            if len(masks[self.chosen_disturb_sensor_config]) == 0:
                self.mode = 'multi'
            elif masks[self.chosen_disturb_sensor_config] == [-1]:
                self.mode = 'dynamic'
                self.dynamic_disturb_time = random.randint(10, 20)

        if self.mode == 'left':
            mask[:occluded_size] = 1
        elif self.mode == 'right':
            mask[occluded_size:] = 1
        elif self.mode == 'center':
            center_left = occluded_size // 2
            center_right = occluded_size - center_left
            center_idx = mask.shape[0] // 2
            mask[: center_idx - center_left] = 1
            mask[center_idx + center_right :] = 1
        elif self.mode == 'multi':
            if not self.turn_on_const_disturb:
                if isinstance(self.config['disturb_mode_multi_num'], list):
                    self.multi_num = random.choice(self.config['disturb_mode_multi_num'])
                else:
                    self.multi_num = self.config['disturb_mode_multi_num']
                probs = self.generate_random_probs(self.multi_num)


                if self.constant_range <= 0.5: # if disturb region out of whole <= 0.5
                    self.divided_num_lst = self.divide_num_with_probs(occluded_size, probs)
                    self.divided_num_lst.sort(reverse=True)
                    self.rand_idx = self.get_mask_idx(self.divided_num_lst)
                    
                    for i in range(self.multi_num):
                        start_idx, mask_length = self.rand_idx[i], self.divided_num_lst[i]
                        mask[start_idx : start_idx + mask_length] = 1
                else:
                    occluded_size = self.config['obs_dim'] - occluded_size
                    self.divided_num_lst = self.divide_num_with_probs(occluded_size, probs)
                    self.divided_num_lst.sort(reverse=True)
                    self.rand_idx = self.get_mask_idx(self.divided_num_lst)
                    
                    mask[:] = 1
                    for i in range(self.multi_num):
                        start_idx, mask_length = self.rand_idx[i], self.divided_num_lst[i]
                        mask[start_idx : start_idx + mask_length] = 0

                self.turn_on_const_disturb = True
            else:
                if self.constant_range <= 0.5:  # if disturb region out of whole <= 0.5
                    for i in range(self.multi_num):
                        start_idx, mask_length = self.rand_idx[i], self.divided_num_lst[i]
                        mask[start_idx : start_idx + mask_length] = 1
                else:
                    mask[:] = 1
                    for i in range(self.multi_num):
                        start_idx, mask_length = self.rand_idx[i], self.divided_num_lst[i]
                        mask[start_idx : start_idx + mask_length] = 0

        elif self.mode == 'spec':
            target_mask = masks[self.chosen_disturb_sensor_config]
            if any(isinstance(i, list) for i in target_mask):
                for m in target_mask:
                    start, end = m
                    mask[start:end] = 1
            else:
                start, end = target_mask
                mask[start:end] = 1

        elif self.mode == 'dynamic':
            # select random transfer time to change current normal sensor config. to disturb one.
            mask[:] = 1
            if self.dynamic_disturb_time == self.timestep:
                if self.config['disturb_enable_dynamic']:
                    self.chosen_disturb_sensor_config = 0
                else:
                    self._update_random_sensor(dynamic=True)
                if len(masks[self.chosen_disturb_sensor_config]) == 0:
                    self.mode = 'multi'
                else:
                    self.mode = 'spec'
        else:
            raise "Error: Please check the disturb mode again!"

        # Apply sensor mask 
        sensor_res = sensor * mask

        # Apply mask data
        mask = sensor * np.logical_not(mask)
        mask = mask.tolist()

        return sensor_res, mask

    def _update_random_sensor(self, dynamic=False):
        if dynamic:
            # exlude last disturb mask, which is dynamic disturb one.
            self.chosen_disturb_sensor_config = random.randint(0, self.num_of_random_sensor_config-3)
        else:
            self.chosen_disturb_sensor_config = random.randint(0, self.num_of_random_sensor_config-1)

    def parse_each_sensor(self, raw_sensor):
        if not isinstance(raw_sensor, np.ndarray):
            raw_sensor = np.array(raw_sensor)
        
        disturbed_sensor_data = []
        for stacked_s in raw_sensor:
            disturbed_stack_s = np.zeros(self.config['stack_N'] * self.config['obs_dim'])
            for stack_index in range(self.config['stack_N']):
                s = stacked_s[stack_index*self.config['obs_dim']:(stack_index+1)*self.config['obs_dim']]
                disturbed_stack_s[stack_index*self.config['obs_dim']:(stack_index+1)*self.config['obs_dim']] = self.trigger_disturbance(s)
            disturbed_sensor_data.append(disturbed_stack_s)

        return disturbed_sensor_data

    def trigger_both_disturbance(self, sensor):
        self.timestep += 1

        if self.config['use_sensor_noise']:
            sensor = self.add_sensor_noise(sensor)

        if self.config['disturb_option'] == 'static':
            sensor, mask = self.trigger_constant_damage(sensor)
        elif self.config['disturb_option'] == 'dynamic':
            sensor = self.trigger_disturbance(sensor)
        elif self.config['disturb_option'] == 'both':
            sensor = self.trigger_disturbance(sensor)
            sensor, mask = self.trigger_constant_damage(sensor)
        else:
            raise Exception("Please set a disturbance option!")
        
        return sensor, mask

    def get_occ_idx(self, obs):
        for i in range(obs.shape[0]):
            if obs[i,0] == self.pcl_filter_val and obs[i,1] == self.pcl_filter_val:
                break
        if i == self.config['obs_dim'] - 1:
            return self.config['obs_dim']
        else:
            return i + 1

    def add_sensor_noise(self, sensor):
        new_s = []
        noise_small = np.random.normal(0, 3, len(sensor)) #/ self.config['sensor_range']
        noise_large = np.random.normal(0, 12, len(sensor)) #/ self.config['sensor_range']
        for i, s in enumerate(sensor):
            if s < 100:
                new_s.append( s + noise_small[i] )
            elif 100 <= s <= 560:
                new_s.append( s + noise_large[i] )
        return new_s

if __name__ == '__main__':
    # Set config
    import yaml
    config = None
    with open('configs.yaml') as f:
        config = yaml.load(f, Loader=yaml.FullLoader)