from mobile2d.lidar_model import LidarModel
from mobile2d.wmr_model import KinematicModel
from mobile2d.utils import *
from envs.disturbed_sensor import DisturbedSensor
from envs.confidence_map import ConfidenceMap
from envs.scenes import get_scene_config
import matplotlib.pyplot as plt
import cv2
import numpy as np
import rvo2


class NavigationEnv:
    def __init__(self, path="maps/map.png", config=None, eval_mode=False):
        # Read Map
        self.img = cv2.flip(cv2.imread(path), 0)
        self.img[self.img>128] = 255
        self.img[self.img<=128] = 0
        self.m = np.asarray(self.img)
        self.m = cv2.cvtColor(self.m, cv2.COLOR_RGB2GRAY)
        self.m = self.m.astype(np.float32) / 255.
        self.m_origin = self.m.copy()
        self.img = self.img.astype(np.float32)/255.
        self.eval_mode = eval_mode

        # Set config
        self.config = config
        
        # Set sensor model
        self.lmodel = LidarModel(self.m, config=self.config)

        # Set boolean last timestep var. 
        self.is_last_timestep = False

        # Set env params
        self.collision_boundary = self.config['collision_boundary']

        # Set confidence map
        if self.config['use_cmap']:
            self.confidence_map = ConfidenceMap(self.m, self.config)
            self.cmap = None
            if self.config['cmap_render']:
                self._set_cmap_plt()

        # Set state_rel distance
        self.diagonal_dist = np.sqrt((self.m.shape[0] ** 2) + (self.m.shape[1] ** 2))

        # Set eval scene given a path
        map_num = int(''.join(filter(str.isdigit, path)))

        # Set disturbed sensor
        self.sensor_disturb = None

        # Initialize ORCA and humans
        self.rvo_sim = None
        self.humans = {}
        self.humans_pos = {}

    def __del__(self):
        if self.config['use_cmap'] and self.config['cmap_render']:
            if hasattr(self, 'fig') and self.fig is not None:
                plt.close(self.fig)

    def reset(self, reset_data=None):
        if self.eval_mode:
            reset_data = self.reset_data

        # Set Mobile Car
        self.car = KinematicModel(v_range=self.config['linear_vel'], w_range=self.config['angular_vel'], \
                                  d=12, wu=12, wv=4, car_w=18, car_f=14, car_r=10, dt=self.config['dt'], config=self.config)
        if self.eval_mode:
            self.car.x, self.car.y, self.car.yaw = reset_data[0]
        else:
            self.car.x, self.car.y = self._search_target()
            self.car.yaw = 360 * np.random.random()
        self.pos = (self.car.x, self.car.y, self.car.yaw)

        # Set Navigation Target
        if self.eval_mode:
            self.target = reset_data[1]
        else:
            while True:
                self.target = self._search_target()
                diff = np.sqrt((self.car.x - self.target[0])**2 + (self.car.y - self.target[1])**2)
                if diff > 150:
                    break

        self.target_dist = np.sqrt((self.car.x - self.target[0])**2 + (self.car.y - self.target[1])**2)
        target_orien = np.arctan2(self.target[1]-self.car.y, self.target[0]-self.car.x) - np.deg2rad(self.car.yaw)
        target_rel = [self.target_dist*np.cos(target_orien), self.target_dist*np.sin(target_orien)]

        # Update image map of sensor model
        self.m = self.m_origin.copy()
        self.lmodel.img_map = self.m
        
        # Initialize humans
        self._initialize_orca()
        self._initialize_humans(num_humans=self.config['num_humans'], reset_data=reset_data)

        # Initialize Measurement
        self.sdata, self.plist = self.lmodel.measure_2d(self.pos)

        if self.config['is_disturbance']:
            self.disturbed_sensor = None
            self.disturbed_sensor = DisturbedSensor(self.config)
            self.sdata, self.disturb_mask = self.disturbed_sensor.trigger_both_disturbance(self.sdata)
            self.sdata, self.plist = self.lmodel.measure_2d(self.pos, self.sdata)

        # Set cmap if True
        if self.config['use_cmap']:
            self.confidence_map.initialize(self.target)

        if self.config['use_polar_coordinate']:
            state = self._construct_state(self.sdata, [self.target_dist, target_orien])  # [distance r, angle theta]
        else:
            state = self._construct_state(self.sdata, target_rel)

        # Get robot's current velocity
        vel_linear, vel_angular = self.car.v, self.car.w

        # Initialize sensor data (used during a episode)
        # At first, stacked sensor data are filled with initial sensor
        self.sensor_data = np.tile(state[:self.config['obs_dim']], self.config['stack_N'])

        pcl_points = self.to_pcl_data(self.sensor_data, self.lmodel.get_theta())

        if self.config['use_cmap']:
            self.orien = np.rad2deg(np.arctan2(self.target[1] - self.car.y, self.target[0] - self.car.x))
            self.cmap = self.confidence_map.step(self.pos, self.plist, self.orien)
            self.prev_cmap = self.cmap
            if self.config['use_recurrent_method']:
                state = pcl_points + state[self.config['obs_dim']:] + [vel_linear/self.config['linear_vel'], vel_angular/self.config['angular_vel']] + self.cmap.tolist()
            else:
                state = self.sensor_data.tolist() + state[self.config['obs_dim']:] + [vel_linear/self.config['linear_vel'], vel_angular/self.config['angular_vel']] + self.cmap.tolist()
        else:
            if self.config['use_recurrent_method']:
                state = pcl_points + state[self.config['obs_dim']:] + [vel_linear/self.config['linear_vel'], vel_angular/self.config['angular_vel']]
            else:
                state = self.sensor_data.tolist() + state[self.config['obs_dim']:] + [vel_linear/self.config['linear_vel'], vel_angular/self.config['angular_vel']]

        return state

    def step(self, action):
        # Update humans
        self.rvo_sim.doStep()
        self._update_humans()

        # Update image map of sensor model
        self.lmodel.img_map = self.m

        # Control and Update robot
        self.car.control((action[0]+1)/2*self.car.v_range, action[1]*self.car.w_range)
        self.car.update()
        collision = self._check_collision()
        
        # Measure Lidar
        self.pos = (self.car.x, self.car.y, self.car.yaw)
        self.sdata, self.plist = self.lmodel.measure_2d(self.pos)
        
        if self.config['is_disturbance']:
            self.sdata, self.disturb_mask = self.disturbed_sensor.trigger_both_disturbance(self.sdata)
            self.sdata, self.plist = self.lmodel.measure_2d(self.pos, self.sdata)

        # Distance Reward
        curr_target_dist = np.sqrt((self.car.x - self.target[0])**2 + (self.car.y - self.target[1])**2)  # L2 distance
        reward_dist = self.target_dist - curr_target_dist

        # Orientation Reward
        self.err_orien = self._compute_rel_diff_angle()
        self.orien = np.rad2deg(np.arctan2(self.target[1] - self.car.y, self.target[0] - self.car.x))

        reward_orien = -1 if abs(self.err_orien) > 120 else 0

        # Action Penalty
        reward_act = -1 if action[0] < -0.5 else 0
        
        # Distance to obs Penalty
        min_range = min(i for i in self.sdata if i > 0)
        min_dist_from_obs = min_range/self.config['sensor_range']
        reward_obs = -1 * (10 - 10*min_dist_from_obs) if min_dist_from_obs < 0.2 else 0

        # Rotation Penalty
        reward_rot = -abs(action[1]) if abs(action[1]) > 0.7 else 0

        if self.config['use_cmap']:
            self.cmap = self.confidence_map.step(self.pos, self.plist, self.orien)

        # Terminal State
        done = 0
        info = {'result': 'none'}
        reward_collision = reward_goal = 0
        if collision:
            reward_collision = self.config['r_collision']
            done = 1
            info['result'] = 'collision'
        elif curr_target_dist < self.config['goal_boundary']:
            reward_goal = self.config['r_arrival']
            done = 1
            info['result'] = 'arrival'
        elif self.is_last_timestep:
            done = 1
            info['result'] = 'timeout'
        
        reward_dist *= self.config['r_distance_weight']
        reward_act *= self.config['r_act_weight']
        reward_orien *= self.config['r_heading_weight']
        reward_obs *= self.config['r_obs_weight']
        reward_rot *= self.config['r_rot_weight']
        reward_base = -1 * self.config['r_base_weight']
        
        if info['result'] == 'arrival' or info['result'] == 'collision':
            reward_dist = 0

        reward_cmap = 0
        reward_cmap_safety = 0
        if self.config['use_cmap']:
            if self.config['use_cmap_reward']:
                cmap_diff = self.cmap - self.prev_cmap
                reward_cmap = self.confidence_map.calc_sensing_val(cmap_diff)
                reward_cmap = max(reward_cmap, 0)
                self.prev_cmap = self.cmap
                reward_cmap = reward_cmap * self.config['r_cmap_weight'] * self.confidence_map.get_dist2goal_ratio()
            
            if self.config['use_cmap_safety_reward']:
                conf_val = self.confidence_map.get_conf_from_cur_pos(self.pos)
                reward_cmap_safety = -self.config['r_cmap_safety_weight'] * (1-conf_val)

        # Total Reward
        reward = reward_goal + reward_collision + reward_dist + reward_orien + reward_base + reward_cmap + reward_cmap_safety

        # Relative Coordinate of Target
        self.target_dist = curr_target_dist
        target_orien = np.arctan2(self.target[1]-self.car.y, self.target[0]-self.car.x) - np.deg2rad(self.car.yaw)
        target_rel = [self.target_dist*np.cos(target_orien), self.target_dist*np.sin(target_orien)] 
        
        if self.config['use_polar_coordinate']:
            state_next = self._construct_state(self.sdata, [self.target_dist, target_orien])  # [distance r, angle theta]
        else:
            state_next = self._construct_state(self.sdata, target_rel)
        
        # Get robot's current velocity
        vel_linear, vel_angular = self.car.v, self.car.w

        # Stack next sensor data
        cur_sensor = np.array(state_next[:self.config['obs_dim']])
        self._stack_sensor(cur_sensor)

        pcl_points = self.to_pcl_data(self.sensor_data, self.lmodel.get_theta())

        if self.config['use_cmap']:
            if self.config['use_recurrent_method']:
                state_next = pcl_points + state_next[self.config['obs_dim']:] + [vel_linear/self.config['linear_vel'], vel_angular/self.config['angular_vel']] + self.cmap.tolist()
            else:
                state_next = self.sensor_data.tolist() + state_next[self.config['obs_dim']:] + [vel_linear/self.config['linear_vel'], vel_angular/self.config['angular_vel']] + self.cmap.tolist()
        else:
            if self.config['use_recurrent_method']:
                state_next = pcl_points + state_next[self.config['obs_dim']:] + [vel_linear/self.config['linear_vel'], vel_angular/self.config['angular_vel']]
            else:
                state_next = self.sensor_data.tolist() + state_next[self.config['obs_dim']:] + [vel_linear/self.config['linear_vel'], vel_angular/self.config['angular_vel']]

        return state_next, reward, done, info
    
    def render(self, render=False):
        img_ = self.img.copy()

        local_map_img_ = None
        if self.config['render_current_local_sensing']:
            local_map_img_ = np.zeros_like(img_)
            for pts in self.plist:
                cv2.line(
                    local_map_img_, 
                    (int(1*self.pos[0]), int(1*self.pos[1])), 
                    (int(1*pts[0]), int(1*pts[1])),
                    (1.0,1.0,1.0), 4, lineType=cv2.LINE_AA)
            local_map_img_ *= 0.6
            local_map_img_ = cv2.flip(local_map_img_, 0)

        for pts in self.plist:
            cv2.line(
                img_, 
                (int(1*self.pos[0]), int(1*self.pos[1])), 
                (int(1*pts[0]), int(1*pts[1])),
                (1.0,0.5,0.0), 1)

        # Draw target_goal
        cv2.circle(img_, (int(self.target[0]), int(self.target[1])), 15, (1.0,0.5,0.7), 3)  

        # Draw the robot
        img_ = self.car.render(img_)

        # Draw humans
        img_ = self._render_humans(img_)

        origin_img = img_.copy()

        img_ = cv2.flip(img_,0)

        # Draw local sensing visualization
        img_ = self._visualize_local_sensing(img_, local_map_img_)

        if self.config['is_visualize_sensor']:
            img_ = self._visualize_sensor(img_)
        img_ = self._visualize_cmap(img_, origin_img)
        if render:
            cv2.imshow("Navigation",img_)
            cv2.waitKey(1)

        return img_.copy()
    
    def _search_target(self):
        im_h, im_w = self.m.shape[0], self.m.shape[1]
        tx = np.random.randint(0,im_w)
        ty = np.random.randint(0,im_h)

        kernel = np.ones((10,10),np.uint8)
        m_dilate = 1 - cv2.dilate(1-self.m, kernel, iterations=3)
        while m_dilate[ty, tx] < 0.5:
            tx = np.random.randint(0,im_w)
            ty = np.random.randint(0,im_h)
        return tx, ty
    
    def _construct_state(self, sensor, target):
        # Normalize sensors
        state_s = []
        for s in sensor:
            normalized_s = s/self.config['sensor_range']
            if normalized_s < self.config['sensor_min_range']:
                normalized_s = 0
            state_s.append(normalized_s)

        if self.config['use_polar_coordinate']:
            state_t_h, state_t_w = self.m.shape
            diagonal = np.sqrt(state_t_h ** 2 + state_t_w ** 2)
            state_t = [target[0] / diagonal, target[1]]
        else:
            state_t_h, state_t_w = self.m.shape
            state_t = [target[0] / state_t_w, target[1] / state_t_h]

        return state_s + state_t

    def _check_collision(self):
        p1,p2,p3,p4 = self.car.car_box
        l1 = Bresenham(p1[0], p2[0], p1[1], p2[1])
        l2 = Bresenham(p1[0], p3[0], p1[1], p3[1])
        l3 = Bresenham(p3[0], p4[0], p3[1], p4[1])
        l4 = Bresenham(p4[0], p2[0], p4[1], p2[1])
        check = l1 + l2 + l3 + l4
        collision = False
        for pts in check:
            try:
                if self.m[int(pts[1]),int(pts[0])] < self.collision_boundary:
                    collision = True
                    self.car.redo()
                    self.car.v = -0.5 * self.car.v
                    break
            except Exception as e:
                print(f'_check_collision() error: {e}')
                
        return collision

    def _compute_rel_diff_angle(self):
        orien = np.rad2deg(np.arctan2(self.target[1] - self.car.y, self.target[0] - self.car.x))
        err_orien = (orien - self.car.yaw) % 360
        if err_orien > 180:
            err_orien = 360 - err_orien
        
        return err_orien

    def _initialize_humans(self, num_humans=4, reset_data=None):
        # print("Initialize humans.")
        if reset_data is not None and reset_data[2] == []:
            return

        for i in range(num_humans):
            self._generate_human(i, reset_data)

    def _generate_human(self, idx, reset_data=None):
        r = self.config['human_radius']
        if self.eval_mode:
            cx, cy = reset_data[2][idx][0]
        else:
            while True:
                cx, cy = self._search_target()
                offset = 50
                if cx - r - offset <= self.car.x <= cx + r + offset and \
                   cy - r - offset <= self.car.y <= cy + r + offset:
                        continue
                else:
                    break
        human = self.rvo_sim.addAgent((cx, cy))

        h, w = self.m.shape[0], self.m.shape[1]
        y, x = np.ogrid[:h, :w]
        dist_from_center = np.sqrt((x - cx)**2 + (y-cy)**2)
        mask = dist_from_center <= r
        self.m[mask] = 0. 

        if self.eval_mode:
            goal_x, goal_y = reset_data[2][idx][1]
        else:
            goal_x, goal_y = self._search_target()
        sign_x, sign_y = goal_x - cx, goal_y - cy

        self.rvo_sim.setAgentPrefVelocity(human, (sign_x*self.config['velocity'], sign_y*self.config['velocity']))
        self.humans[idx] = {'instance': human, 'r': r, 'mask': mask, 'goal': (goal_x, goal_y)}
        self.humans_pos[idx] = (cx, cy)

    def _render_humans(self, img):
        b,g,r = self.config['agent_color_b'], self.config['agent_color_g'], self.config['agent_color_r']
        color = (b/255, g/255, r/255)
        for _, human in self.humans.items():
            inst = human['instance']
            radius = human['r']
            x, y = self.rvo_sim.getAgentPosition(inst)
            x, y = int(x), int(y)
            img = cv2.circle(img, (x,y), radius, color, -1)  # -1 means to fill in the circle

        return img

    def _update_humans(self):
        for k, human in self.humans.items():
            inst = human['instance']
            mask = human['mask']
            r = human['r']
            self.m[mask] = 1.

            cx, cy = self.rvo_sim.getAgentPosition(inst)
            h, w = self.m.shape[0], self.m.shape[1]
            y, x = np.ogrid[:h, :w]
            dist_from_center = np.sqrt((x - cx)**2 + (y-cy)**2)
            mask_next = dist_from_center <= r
    
            self.m[mask_next] = 0.
            self.humans[k]['mask'] = mask_next

            goal_x, goal_y = human['goal']

            if np.sqrt((cx - goal_x)**2 + (cy - goal_y)**2) < 40: # when arrives at the goal,
                if self.eval_mode:
                    self.rvo_sim.setAgentPrefVelocity(inst, (0, 0))
                else:
                    new_goal_x, new_goal_y = self._search_target()
                    self.humans[k]['goal'] = (new_goal_x, new_goal_y)
                    sign_x, sign_y = new_goal_x - cx, new_goal_y - cy

                    self.rvo_sim.setAgentPrefVelocity(inst, (sign_x*self.config['velocity'], sign_y*self.config['velocity']))

    def _initialize_orca(self):
        # Initialize rvo sim.
        self.rvo_sim = rvo2.PyRVOSimulator(self.config['dt'], self.config['neighbor_dist'], \
                                            self.config['max_neighbors'], self.config['time_horizon'],\
                                            self.config['time_horizon_obs'], self.config['human_radius'],\
                                            self.config['max_speed'], (self.config['velocity'], self.config['velocity']))
        
        imgray = cv2.cvtColor(self.img, cv2.COLOR_BGR2GRAY)
        _, thr = cv2.threshold(imgray, 0.3, 1, cv2.THRESH_BINARY)
        thr = thr.astype(np.uint8)
        contours, _ = cv2.findContours(thr, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        
        obs_list = []
        for line_seg in contours:
            obs_boundary = []
            for point in line_seg:
                x, y = point[0]
                obs_boundary.append((x, y))
            obs_list.append(obs_boundary)
        for obs in obs_list:
            self.rvo_sim.addObstacle(obs)
        self.rvo_sim.processObstacles()
    
    def _visualize_sensor(self, origin_img):
        origin_H, origin_W = origin_img.shape[0], origin_img.shape[1]
        sensor_viz_H, sensor_viz_W = origin_H // 4, origin_W

        sensor_img = self.sensor_data[::-1].reshape(self.config['stack_N'], self.config['obs_dim']).astype(np.float32)
        sensor_img = cv2.resize(sensor_img, (sensor_viz_W, sensor_viz_H))
        sensor_img = cv2.cvtColor(sensor_img, cv2.COLOR_GRAY2BGR)
        sensor_img = cv2.flip(sensor_img, 0)
        
        concat_img = np.concatenate((sensor_img, origin_img), axis=0)

        return concat_img

    def _stack_sensor(self, cur_sensor):
        self.sensor_data[self.config['obs_dim']:] = self.sensor_data[:-self.config['obs_dim']]
        self.sensor_data[:self.config['obs_dim']] = cur_sensor

    def _render_cmap(self, img_):
        color = (0/255., 255/255., 204/255.)
        img_ = cv2.circle(img_, (int(round(self.car.x)), int(round(self.car.y))), self.config['cmap_radius'], color, 1)

        return img_

    def _set_cmap_plt(self):
        self.fig = plt.figure()
        self.plt_x = np.arange(self.config['cmap_sector_num'])
        self.plt_y = np.zeros(self.config['cmap_sector_num'])
        self.plt_line, = plt.plot(self.plt_x, self.plt_y, 'b-')
        plt.ylim(-0.1, 1.1)

    def _visualize_local_sensing(self, img_, local_map_img_):
        if local_map_img_ is None:
            local_map_img_ = self.confidence_map.m
            local_map_img_ = cv2.merge([local_map_img_, local_map_img_, local_map_img_])
            local_map_img_ = cv2.flip(local_map_img_, 0)
            img_ = 0.4 * img_ + local_map_img_ * 0.6
        else:
            img_ = 0.4 * img_ + local_map_img_

        return img_

    def _visualize_cmap(self, img_, origin_img):
        origin_img = cv2.flip(origin_img, 0) # flip back

        origin_h, origin_w = img_.shape[0], img_.shape[1]
        c_img_h, c_img_w = origin_h // 2, origin_w // 2
        vc_img_h, vc_img_w = origin_h - c_img_h, origin_w // 2
        
        if self.config['cmap_render']:
            conf_val_img_h, conf_val_img_w = 30, vc_img_w
            vc_img_h = origin_h - c_img_h - conf_val_img_h

            self.plt_line.set_ydata(self.cmap)
            self.fig.canvas.draw()
            
            vc_img = np.frombuffer(self.fig.canvas.tostring_rgb(), dtype=np.uint8)
            vc_img  = vc_img.reshape(self.fig.canvas.get_width_height()[::-1] + (3,))

            vc_img = cv2.cvtColor(vc_img,cv2.COLOR_RGB2BGR)
            vc_img = cv2.resize(vc_img, (vc_img_w, vc_img_h), cv2.INTER_NEAREST) / 255.

            cmap_sum = self.confidence_map.calc_sensing_val(self.cmap)
            cmap_sum /= self.config['cmap_sector_num']
            cmap_sum *= 100
            cmap_sum = round(cmap_sum, 2)
            if cmap_sum == 100:
                cmap_sum = '100'
            else:
                cmap_sum = str(cmap_sum)
            conf_val = 'Value: ' + cmap_sum + '%'
            conf_val_img = np.ones((conf_val_img_h, conf_val_img_w, 3))
            conf_val_img = cv2.putText(conf_val_img, conf_val, (conf_val_img.shape[1]//4, conf_val_img.shape[0]-conf_val_img.shape[0]//15), 0, 0.5, (0,0,0), 1)
        else:
            vc_img = np.ones((vc_img_h, vc_img_w, 3))

        origin_img = cv2.resize(origin_img, (c_img_w, c_img_h), cv2.INTER_NEAREST)
        if self.config['cmap_render']:
            cmap_img = np.concatenate((origin_img, conf_val_img, vc_img), axis=0).astype(np.float32)
        else:
            cmap_img = np.concatenate((origin_img, vc_img), axis=0).astype(np.float32)

        dummy_w = 20
        dummy_img = np.ones((origin_h, dummy_w, 3))
        img_ = np.concatenate((img_, dummy_img, cmap_img, dummy_img), axis=1)

        return img_

    def polar2cart(self, radius, theta):
        x = radius * np.cos(theta)
        y = radius * np.sin(theta)
        return x, y

    def to_pcl_data(self, sensors, theta):
        points = []
        mask_points = []
        theta = np.deg2rad(theta)
        for s, theta in zip(sensors, theta):
            if s != 0:
                x, y = self.polar2cart(s, theta)
                points += [x, y]
            else:
                mask_points += [self.config['pcl_filter_val'], self.config['pcl_filter_val']]

        return points + mask_points