import torch
import random
from collections import deque, namedtuple

Transition = namedtuple('Transition', ('state', 'action', 'reward', 'next_state', 'done', 'hidden'))

class ReplayBufferGRU:
    def __init__(self, config):
        self.config = config
        self.capacity = self.config['memory_size']
        self.buffer = deque(maxlen=self.capacity)
        self.sequence_length = self.config['recurrent_length']
        self.batch_size = self.config['batch_size']
        self.local_buffer = []

    def push(self, state, action, reward, next_state, done, hidden):
        self.local_buffer.append(Transition(state, action, reward, next_state, done, hidden.squeeze(1)))
        
        if done == 1:
            if len(self.local_buffer) >= self.sequence_length:
                self.buffer.append(self.local_buffer)
                self.local_buffer = []
            else:
                self.local_buffer = []
                
    def sample(self):
        s_lst, a_lst, r_lst, ns_lst, d_lst, h_lst = [],[],[],[],[],[]
        batch = random.sample(self.buffer, self.batch_size)
        for sample in batch:
            start = random.randint(0, len(sample) - self.sequence_length)
            transitions = sample[start : start + self.sequence_length]
            batch = Transition(*zip(*transitions))

            s_lst.append(torch.FloatTensor(batch.state))  # [seq_len, input_dim]
            a_lst.append(torch.FloatTensor(batch.action))
            r_lst.append(torch.FloatTensor(batch.reward))
            ns_lst.append(torch.FloatTensor(batch.next_state))
            d_lst.append(torch.FloatTensor(batch.done))
            h_lst.append(torch.stack(list(batch.hidden)))  # [seq_len, 1(gru), input_dim]

        h_lst = torch.cat(h_lst, dim=-2).detach()   # cat along the batch dim

        return s_lst, a_lst, r_lst, ns_lst, d_lst, h_lst

    def __len__(self):
        return len(self.buffer)

    def get_length(self):
        return len(self.buffer)