import sac
import conv_model_pcl_gru
import numpy as np
import os
import rl_eval
import yaml
import datetime
import time
import ray
import torch
import shutil

from envs.env import NavigationEnv
from torch.utils.tensorboard import SummaryWriter


def set_agent(config):
    batch_size = config['batch_size']
    memory_size = config['memory_size']
    reward_decay = config['reward_decay']
    actor_lr = config['actor_lr']
    critic_lr = config['critic_lr']
    agent = None

    actor_model = conv_model_pcl_gru.PolicyNetGaussian
    if config['use_double_q_net']:
        critic_model = conv_model_pcl_gru.DoubleQNet
    else:
        critic_model = conv_model_pcl_gru.QNet
    
    agent = sac.SAC(
        model = [actor_model, critic_model],
        n_actions = 2,
        learning_rate = [actor_lr, critic_lr],
        reward_decay = reward_decay,
        memory_size = memory_size,
        batch_size = batch_size,
        alpha = config['sac_alpha'],
        config = config,
        auto_entropy_tuning=True,
        entropy_lr=config['entropy_lr'],
        beta=config['cmap_confidence_coefficient'])

    return agent

def load_epnet(agent, ep_net_path):
    agent.load_epnet(ep_net_path)
    for param in agent.epnet.parameters():
        param.requires_grad = False

def load_trained_model_to_agent(agent, config):
    model_path = config['resume_path']
    print("Load model ...", model_path)
    file_dir = os.listdir(model_path)
    last_epoch = [int(f.split('_')[-1][:-3]) for f in file_dir]
    last_epoch = max(last_epoch)

    # load model
    anet_path = model_path + config['algo_name'] + "_anet_"+ str(last_epoch) + ".pt"
    cnet_path = model_path + config['algo_name'] + "_cnet_"+ str(last_epoch) + ".pt"
    alpha_path = model_path + config['algo_name'] + "_alpha_"+ str(last_epoch) + ".pt"
    agent.save_load_model("load", anet_path, cnet_path, epnet_path=False, alpha_path=alpha_path, lazy_load_epnet=True)

    # load optim.
    actor_optim_path = model_path + config['algo_name'] + "_actor_optim_"+ str(last_epoch) + ".pt"
    critic_optim_path = model_path + config['algo_name'] + "_critic_optim_"+ str(last_epoch) + ".pt"
    alpha_optim_path = model_path + config['algo_name'] + "_alpha_optim_"+ str(last_epoch) + ".pt"
    agent.save_load_optim("load", actor_optim_path, critic_optim_path, alpha_optim_path)

    start_epi = last_epoch

    return start_epi

def main(config):
    #######################################################################
    #                              configs
    #######################################################################
    batch_size = config['batch_size']
    max_epi_len =  config['max_epi_len']
    epi_per_eval = config['epi_per_eval']
    total_itr = config['total_itr']
    render = config['render']
    load_model = config['load_model']
    map_path = config['map_path']
    exp_name = config['exp_name']
    model_path = config['model_path'] + exp_name
    gif_path = config['gif_path'] + exp_name
    total_epi_for_eval = config['total_epi_for_eval']
    log_path = config['log_path']

    if config['use_curriculum_learning']:
        config['is_disturbance'] = False
        config['use_cmap'] = False
        config['cmap_render'] = False
        config['render_current_local_sensing'] = True
            
    # tensorboard
    now = datetime.datetime.now()
    date_time = now.strftime('%Y-%m-%d')
    writer = SummaryWriter(log_path + date_time + '/' + exp_name)

    # save configs
    shutil.copy('./configs.yaml', log_path + date_time + '/'+ exp_name + '/')

    # Set rl algorithms
    agent = set_agent(config)

    # Create directory
    if not os.path.exists(model_path):
        os.makedirs(model_path)

    # Load pre-trained rl agent
    if load_model:
        start_epi = load_trained_model_to_agent(agent, config)
    else:
        start_epi = 0

    # ray setting
    n_cpu = os.cpu_count()
    print(f'Currently used cpu: {n_cpu}')
    ray.init(num_cpus=n_cpu)

    if config['use_cmap']:
        if config['load_epnet']:
            load_epnet(agent, ep_net_path=config['epnet_path'])

    if config['use_random_map']:
        total_map_path_list = [os.path.join(config['map_root_path'], i) for i in os.listdir(config['map_root_path'])]
        total_map_num = len(total_map_path_list)

    total_step = 0
    success_count = 0
    collision_count = 0
    timeout_count = 0
    init_all_config = False

    #######################################################################
    #                              Training
    #######################################################################
    start_time = time.time()
    for eps in range(start_epi, total_itr):
        if config['use_curriculum_learning']:
            if total_step > config['start_step_curriculum_learning']:
                config['is_disturbance'] = True
                config['use_cmap'] = True
                config['cmap_render'] = True
                config['render_current_local_sensing'] = False

                if not init_all_config:
                    agent.init_memory()
                    init_all_config = True
                    agent = set_agent(config)

                    if config['use_cmap']:
                        if config['load_epnet']:
                            load_epnet(agent)
                    env = NavigationEnv(path=total_map_path_list[map_idx], config=config)
                    
        if config['use_random_map']:
            if eps % config['random_map_per_epi'] == 0:
                map_idx = (eps // config['random_map_per_epi']) % total_map_num 
                env = NavigationEnv(path=total_map_path_list[map_idx], config=config)
        state = env.reset()
        step = 0
        loss_a = loss_c = 0.
        acc_reward = 0.
        
        hidden = torch.zeros((1, 1, config['recurrent_hidden_dim']))

        #######################################################################
        #                              Start episode
        #######################################################################
        while True:
            if config['init_exploration_step'] > total_step and not load_model:
                _, hidden_next = agent.choose_action(state, hidden, eval=False)
                action = np.random.uniform(-1,1,2)
            else:
                action, hidden_next = agent.choose_action(state, hidden, eval=False)

            # Check whether the current step is the last timestep
            last_timestep = True if step >= (max_epi_len - 1) else False
            env.is_last_timestep = last_timestep
            state_next, reward, done, info = env.step(action)

            agent.store_recurrent_transition(state, action, reward, state_next, done, hidden)

            if render:
                env.render(render=render)

            # Learn the model
            loss_a = loss_c = 0.
            if total_step > batch_size and total_step > config['start_training_step']:
                if len(agent.buffer) > config['batch_size']:
                    loss_a, loss_c = agent.learn()
            
            step += 1
            total_step += 1
            acc_reward += reward
            state = state_next.copy()
            hidden = hidden_next

            writer_step = step + max_epi_len * eps
            writer.add_scalar('Training/Actor_loss', loss_a, writer_step)
            writer.add_scalar('Training/Critic_loss', loss_c, writer_step)
            writer.add_scalar('Training/alpha', agent.alpha, writer_step)
            print('\rEps:{:3d} /{:4d} /{:6d}| action:{:+.2f}| R:{:+.2f}| Loss:[A>{:+.2f} C>{:+.2f}]| Alpha: {:.3f}| Ravg:{:.2f}'\
                    .format(eps, step, total_step, action[0], reward, loss_a, loss_c, agent.alpha, acc_reward/step), end='')
            if config['use_cmap']:
                writer.add_scalar('Training/Sensing_confidence_per_step', env.confidence_map.sensed_val, writer_step)

            if done:
                if info['result'] == 'arrival':
                    success_count += 1
                elif info['result'] == 'collision':
                    collision_count += 1
                elif info['result'] == 'timeout':
                    timeout_count += 1
                print('| S/C/T: [{}/{}/{}]'.format(success_count, collision_count, timeout_count))
                break
                
        #######################################################################
        #                              End episode
        #######################################################################
        writer.add_scalar('Training/Avg_reward_per_episode', acc_reward/step, eps)
        writer.add_scalar('Training/Sum_of_reward_per_episode', acc_reward, eps)

        if config['use_cmap']:
            sum_of_sensing_conf = env.confidence_map.get_total_sensing_conf()
            writer.add_scalar('Training/Total_sensing_confidence_per_episode', sum_of_sensing_conf, eps)

        if (eps > 0) and (eps % epi_per_eval == 0) and total_step >= config['init_exploration_step']-1:
            #######################################################################
            #                              Evaluation
            #######################################################################
            success_rate = success_count / epi_per_eval
            success_count = 0
            collision_rate = collision_count / epi_per_eval
            collision_count = 0
            timeout_rate = timeout_count / epi_per_eval
            timeout_count = 0

            writer.add_scalar('Training/Success_Rate', success_rate, eps)
            writer.add_scalar('Training/Collision_Rate', collision_rate, eps)
            writer.add_scalar('Training/Timeout_Rate', timeout_rate, eps)

            # Save the model
            print("Save model to " + model_path)
            anet_path = model_path + config['algo_name'] + "_anet_"+ str(eps) + ".pt"
            cnet_path = model_path + config['algo_name'] + "_cnet_"+ str(eps) + ".pt"
            epnet_path = model_path + config['algo_name'] + "_epnet_"+ str(eps) + ".pt"
            alpha_path = model_path + config['algo_name'] + "_alpha_"+ str(eps) + ".pt"
            agent.save_load_model("save", anet_path, cnet_path, epnet_path, alpha_path)

            actor_optim_path = model_path + config['algo_name'] + "_actor_optim_"+ str(eps) + ".pt"
            critic_optim_path = model_path + config['algo_name'] + "_critic_optim_"+ str(eps) + ".pt"
            alpha_optim_path = model_path + config['algo_name'] + "_alpha_optim_"+ str(eps) + ".pt"
            agent.save_load_optim("save", actor_optim_path, critic_optim_path, alpha_optim_path)
            
            # Print Success Rate and Elapsed time
            sec = time.time() - start_time
            elapsed_time = str(datetime.timedelta(seconds=sec)).split('.')[0]
            print("Elapsed time: {}".format(elapsed_time)) 
            
            print("Evaluating...")
            gif_ = map_path[5:].split('.')[0] + "_" + str(eps).zfill(7) + ".gif"

            # Run evaluation
            avg_rewards, success_rate, collision_rate, timeout_rate, avg_path_length, avg_arrival_time = \
                rl_eval.run(agent, total_eps=total_epi_for_eval, map_path=map_path, gif_path=gif_path, \
                            gif_name=gif_, config=config, use_random_map=config['use_random_map'], log_path=None, extract_video=False)

            print("Evaluation done!")
            writer.add_scalar('Testing/Avg_reward', avg_rewards, eps)
            writer.add_scalar('Testing/Success Rate', success_rate, eps)
            writer.add_scalar('Testing/Collision rate', collision_rate, eps)
            writer.add_scalar('Testing/Timeout rate', timeout_rate, eps)
            writer.add_scalar('Testing/Avg path length', avg_path_length, eps)
            writer.add_scalar('Testing/Avg arrival time', avg_arrival_time, eps)

    # End training
    writer.close()


if __name__ == "__main__":
    with open('configs.yaml') as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
    main(config)