import sys
sys.path.append("./mobile2d")
import numpy as np
import cv2
import sys
import yaml
import time
import ray
import os
# from utils import *


class LidarModel:
    def __init__(self,
            img_map,
            sensor_size = 64,
            start_angle = -120.0,
            end_angle = 120.0,
            max_dist = 200.0,
            config = None
        ):
        self.config = config
        self.sensor_size = self.config['obs_dim']  # Same as sensor_dim
        self.start_angle = self.config['sensor_angle_min']
        self.end_angle = self.config['sensor_angle_max']
        self.max_dist = self.config['sensor_range']
        self.inter = (self.end_angle-self.start_angle) / (self.sensor_size-1)
        self.img_map = img_map
        self.theta_lst = []
    

    def get_theta(self):
        return self.theta_lst

    def measure_parallel(self, pos):
        img_map = ray.put(self.img_map)
        max_dist = ray.put(self.max_dist)

        sense_data = []
        self.theta_lst = []
        for i in range(self.sensor_size):
            theta = i * self.inter
            sense_data.append( ray_cast.remote(img_map,
                                      max_dist,
                                      np.array((pos[0], pos[1])), 
                                      pos[2] + self.start_angle + theta
                                      ) )
            self.theta_lst.append(theta)

        sense_data = ray.get(sense_data)

        return sense_data
    
    def measure_2d(self, pos, disturb_sdata=None):
        if disturb_sdata is not None:
            plist = EndPoint(pos, [self.sensor_size, self.start_angle, self.end_angle], disturb_sdata)
            return disturb_sdata, plist
        else:
            # sdata = self.measure(pos)
            sdata = self.measure_parallel(pos)
            plist = EndPoint(pos, [self.sensor_size, self.start_angle, self.end_angle], sdata)
            return sdata, plist

@ ray.remote
def ray_cast(img_map, max_dist, pos, theta):
    end = np.array((pos[0] + max_dist*np.cos(np.deg2rad(theta)), pos[1] + max_dist*np.sin(np.deg2rad(theta))))
    x0, y0 = int(pos[0]), int(pos[1])
    x1, y1 = int(end[0]), int(end[1])
    plist = Bresenham(x0, x1, y0, y1)
    dist = max_dist
    for p in plist:
        if p[1] >= img_map.shape[0] or p[0] >= img_map.shape[1] or p[1]<0 or p[0]<0:
            continue
        if img_map[p[1], p[0]] < 0.5:
            tmp = np.power(float(p[0]) - pos[0], 2) + np.power(float(p[1]) - pos[1], 2)
            tmp = np.sqrt(tmp)
            if tmp < dist:
                dist = tmp
    return dist


def Bresenham(x0, x1, y0, y1):
    rec = []
    dx = abs(x1 - x0)
    dy = abs(y1 - y0)
    x, y = x0, y0
    sx = -1 if x0 > x1 else 1
    sy = -1 if y0 > y1 else 1
    if dx > dy:
        err = dx / 2.0
        while x != x1:
            rec.append((x, y))
            err -= dy
            if err < 0:
                y += sy
                err += dx
            x += sx
    else:
        err = dy / 2.0
        while y != y1:
            rec.append((x, y))
            err -= dx
            if err < 0:
                x += sx
                err += dy
            y += sy
    return rec

def EndPoint(pos, bot_param, sensor_data):
    pts_list = []
    inter = (bot_param[2] - bot_param[1]) / (bot_param[0]-1)
    for i in range(bot_param[0]):
        theta = pos[2] + bot_param[1] + i*inter
        pts_list.append(
            [ pos[0]+sensor_data[i]*np.cos(np.deg2rad(theta)),
              pos[1]+sensor_data[i]*np.sin(np.deg2rad(theta))] )
    return pts_list

def gaussian(x, mu, sig):
    return 1./(np.sqrt(2.*np.pi)*sig)*np.exp(-np.power((x - mu)/sig, 2.)/2)


if __name__ == "__main__":
    # ray setting
    n_cpu = os.cpu_count()
    print(f'current used cpu: {n_cpu}')
    ray.init(num_cpus=n_cpu)

    img = cv2.flip(cv2.imread("maps/map.png"),0)
    img[img>128] = 255
    img[img<=128] = 0
    m = np.asarray(img)
    m = cv2.cvtColor(m, cv2.COLOR_RGB2GRAY)
    m = m.astype(float) / 255.
    img = img.astype(float)/255.

    config = None
    with open('configs.yaml') as f:
        config = yaml.load(f, Loader=yaml.FullLoader)

    lmodel = LidarModel(m, config=config)
    pos = (100,200,0)

    while True:
        start_time = time.time()
        sdata, plist = lmodel.measure_2d(pos)
        end_time = time.time()
        print(f'measure time: {end_time - start_time}')
        img_ = img.copy()
        for pts in plist:
            cv2.line(
                img_, 
                (int(1*pos[0]), int(1*pos[1])), 
                (int(1*pts[0]), int(1*pts[1])),
                (0.0,1.0,0.0), 1)
        cv2.circle(img_,(pos[0],pos[1]),5,(0.5,0.5,0.5),3)
        img_ = cv2.flip(img_,0)
        cv2.imshow("Lidar Test", img_)
        k = cv2.waitKey(0)

        if k == 27:  # ESC
            print()
            break