import cv2
import numpy as np

def _rot_pos(x,y,phi_):
    phi = np.deg2rad(phi_)
    return np.array((x*np.cos(phi)+y*np.sin(phi), -x*np.sin(phi)+y*np.cos(phi)))

def _draw_rectangle(img,x,y,u,v,phi,color=(0,0,0),size=1):
    pts1 = _rot_pos(-u/2,-v/2,phi) + np.array((x,y))
    pts2 = _rot_pos(u/2,-v/2,phi) + np.array((x,y))
    pts3 = _rot_pos(-u/2,v/2,phi) + np.array((x,y))
    pts4 = _rot_pos(u/2,v/2,phi) + np.array((x,y))
    cv2.line(img, tuple(pts1.astype(np.int).tolist()), tuple(pts2.astype(np.int).tolist()), color, size)
    cv2.line(img, tuple(pts1.astype(np.int).tolist()), tuple(pts3.astype(np.int).tolist()), color, size)
    cv2.line(img, tuple(pts3.astype(np.int).tolist()), tuple(pts4.astype(np.int).tolist()), color, size)
    cv2.line(img, tuple(pts2.astype(np.int).tolist()), tuple(pts4.astype(np.int).tolist()), color, size)
    return img

class KinematicModel:
    def __init__(self,
            v_range = 30,
            w_range = 60,
            d = 14,
            wu = 10,
            wv = 4,
            car_w = 44,
            car_f = 27,
            car_r = 27,
            dt = 0.1,
            record_max = 10000,
            record_track=False,
            config=None
        ):
        self.init_state((0,0,0))

        self.config = config

        self.v_range = v_range
        self.w_range = w_range
        self.d = d
        self.wu = wu
        self.wv = wv
        self.car_w = car_w
        self.car_f = car_f
        self.car_r = car_r
        self._compute_car_box((self.x, self.y, self.yaw))
        self.dt = dt
        self.rec_max = record_max

        self.record_track = self.config['record_track']
        if self.record_track:
            self.record_interval_step = self.config['record_interval_step']
            self.pose_list = []
    
    def init_state(self,pos):
        self.x = pos[0]
        self.y = pos[1]
        self.yaw = pos[2]
        self.v = 0
        self.a = 0
        self.w = 0
        self.record = []
        self.timestep = 0

    def update(self):
        if self.v > self.v_range:
            self.v = self.v_range
        elif self.v < -self.v_range:
            self.v = -self.v_range
        if self.w > self.w_range:
            self.w = self.w_range
        elif self.w < -self.w_range:
            self.w = -self.w_range
            
        self.x += self.v * np.cos(np.deg2rad(self.yaw)) * self.dt #* self.config['ratio']
        self.y += self.v * np.sin(np.deg2rad(self.yaw)) * self.dt #* self.config['ratio']
        self.yaw += self.w * self.dt
        self.yaw = self.yaw % 360
        self.record.append((self.x, self.y, self.yaw))
        self._compute_car_box((self.x, self.y, self.yaw))

        if self.record_track:
            self.timestep += 1
            if self.timestep % self.record_interval_step == 0:
                self.pose_list.append((self.x, self.y, self.yaw))
    
    def redo(self):
        self.x -= self.v * np.cos(np.deg2rad(self.yaw)) * self.dt #* self.config['ratio']
        self.y -= self.v * np.sin(np.deg2rad(self.yaw)) * self.dt #* self.config['ratio']
        self.yaw -= self.w * self.dt
        self.yaw = self.yaw % 360
        self.record.pop()

    def control(self,v,w):
        self.v = v
        self.w = w

        if self.v > self.v_range:
            self.v = self.v_range
        elif self.v < -self.v_range:
            self.v = -self.v_range
        if self.w > self.w_range:
            self.w = self.w_range
        elif self.w < -self.w_range:
            self.w = -self.w_range

    def state_str(self):
        return "x={:.4f}, y={:.4f}, v={:.4f}, a={:.4f}, yaw={:.4f}, w={:.4f}".format(self.x, self.y, self.v, self.a, self.yaw, self.w)

    def _compute_car_box(self, pose):
        x, y, yaw = pose
        pts1 = _rot_pos(self.car_f,self.car_w/2,-yaw) + np.array((x,y))
        pts2 = _rot_pos(self.car_f,-self.car_w/2,-yaw) + np.array((x,y))
        pts3 = _rot_pos(-self.car_r,self.car_w/2,-yaw) + np.array((x,y))
        pts4 = _rot_pos(-self.car_r,-self.car_w/2,-yaw) + np.array((x,y))
        car_box = (pts1.astype(int), pts2.astype(int), pts3.astype(int), pts4.astype(int))
        self.car_box = car_box

        return car_box

    def get_path_length(self):
        res_len = 0.0
        # start = 0 if len(self.record) < self.rec_max else len(self.record) - self.rec_max
        for i in range(len(self.record)-1):
            t = np.array([self.record[i][0], self.record[i][1]])
            t_n = np.array([self.record[i+1][0], self.record[i+1][1]])
            res_len += np.linalg.norm(t-t_n)

        return res_len

    def render(self, img=np.ones((600,600,3))):
        start = 0 if len(self.record) < self.rec_max else len(self.record) - self.rec_max
        color = (0/255,97/255,255/255)
        for i in range(start,len(self.record)-1):
            cv2.line(img,(int(self.record[i][0]),int(self.record[i][1])), (int(self.record[i+1][0]),int(self.record[i+1][1])), color, 3)

        cur_pose = (self.x, self.y, self.yaw)
        img = self.render_car(img, self._compute_car_box(cur_pose), cur_pose)

        if self.record_track:
            img = self.render_track(img)

        return img

    def render_car(self, img, car_box, pose):
        x, y, yaw = pose
        # Car box
        pts1, pts2, pts3, pts4 = car_box
        color = (0,0,0)
        size = 1
        cv2.line(img, tuple(pts1.astype(np.int).tolist()), tuple(pts2.astype(np.int).tolist()), color, size)
        cv2.line(img, tuple(pts1.astype(np.int).tolist()), tuple(pts3.astype(np.int).tolist()), color, size)
        cv2.line(img, tuple(pts3.astype(np.int).tolist()), tuple(pts4.astype(np.int).tolist()), color, size)
        cv2.line(img, tuple(pts2.astype(np.int).tolist()), tuple(pts4.astype(np.int).tolist()), color, size)
        # Car center & direction
        t1 = _rot_pos( 6, 0, -yaw) + np.array((x,y))
        t2 = _rot_pos( 0, 4, -yaw) + np.array((x,y))
        t3 = _rot_pos( 0, -4, -yaw) + np.array((x,y))
        cv2.line(img, (int(x),int(y)), (int(t1[0]), int(t1[1])), (0,0,1), 2)
        cv2.line(img, (int(t2[0]), int(t2[1])), (int(t3[0]), int(t3[1])), (1,0,0), 2)
        ########## Draw Wheels ##########
        w1 = _rot_pos( 0, self.d, -yaw) + np.array((x,y))
        w2 = _rot_pos( 0,-self.d, -yaw) + np.array((x,y))
        # 4 Wheels
        img = _draw_rectangle(img,int(w1[0]),int(w1[1]),self.wu,self.wv,-yaw)
        img = _draw_rectangle(img,int(w2[0]),int(w2[1]),self.wu,self.wv,-yaw)
        # Axle
        img = cv2.line(img, tuple(w1.astype(np.int).tolist()), tuple(w2.astype(np.int).tolist()), (0,0,0), 1)
        
        # ########## Draw direction ##########
        # arrow_pos = _rot_pos( 20,0, -self.yaw) + np.array((self.x,self.y))
        # img = cv2.arrowedLine(img, (int(self.x),int(self.y)), (int(arrow_pos[0]), int(arrow_pos[1])), (0.2,0.3,1), 2, tipLength=0.3)

        return img

    def render_track(self, img):
        for pose in self.pose_list:
            self.render_car(img, self._compute_car_box(pose), pose)

        return img

# ================= main =================
if __name__ == "__main__":
    # Set config
    import yaml
    config = None
    with open('configs.yaml') as f:
        config = yaml.load(f, Loader=yaml.FullLoader)

    car = KinematicModel(config=config)
    car.init_state((300,300,0))
    step = 0
    while True:
        print("\r[step:{}] x={}, y={}, v={}, yaw={}, w={}".format(step,str(car.x)[:5],str(car.y)[:5],str(car.v)[:5],str(car.yaw)[:5],str(car.w)[:5]), end="\t")
        img = np.ones((600,600,3))
        car.update()
        print(f'car: {car._compute_car_box((car.x,car.y,car.yaw))}')
        img = car.render(img)
        img = cv2.flip(img, 0)
        cv2.imshow("demo", img)
        k = cv2.waitKey(1)

        if k == ord("a"):
            car.w += 5
        elif k == ord("d"):
            car.w -= 5
        elif k == ord("w"):
            car.v += 4
        elif k == ord("s"):
            car.v -= 4
        elif k == 27:  # ESC
            print()
            break
        step += 1

