from envs.env import NavigationEnv
import conv_model
import numpy as np
import os
from PIL import Image
import cv2
import time
import sac
import yaml
import eval_utils
import ray
import random
import torch

def run(rl_core, total_eps=2, message=False, render=False, \
        map_path="maps/map.png", gif_path="eval_gif/", \
        gif_name="test.gif", config=None, eval_filename=False, \
        eval_mode=False, use_random_map=False, log_path=None, extract_video=False):

    if not os.path.exists(gif_path):
        os.makedirs(gif_path)

    if use_random_map:
        total_map_path_list = sorted([os.path.join(config['map_root_path'], i) for i in os.listdir(config['map_root_path'])])
        total_map_num = len(total_map_path_list)
    else:
        env = NavigationEnv(path=map_path, config=config, eval_mode=eval_mode)

    images = []
    avg_reward_list = []

    success_epi_list = []
    collision_epi_list = []
    timeout_epi_list = []

    total_res_list = []

    success_count = 0
    collision_count = 0
    timeout_count = 0
    record_len = total_map_num
    record_count = 1

    path_length = []
    arrival_time = []


    if eval_mode:
        data_list = eval_utils.load_eval_data(eval_filename)

    rl_core.actor.eval()
    for eps in range(total_eps):
        if eval_mode:
            seed_everything(config['random_seed'] + eps)

        if extract_video:
            asked = False
            if eps in [11,16,18,58,66,73,71,84,124,142,147,171,202,208,242,248]: # arbit-disturb
                render = True
            else:
                render = False
                continue

        if use_random_map and eval_mode:
            eps_num_per_map = total_eps // total_map_num
            map_idx = eps // eps_num_per_map
            env = NavigationEnv(path=total_map_path_list[map_idx], config=config)
        
        else:
            map_idx = eps % total_map_num
            env = NavigationEnv(path=total_map_path_list[map_idx], config=config, eval_mode=eval_mode)
        
        step = 0
        acc_reward = 0.
        
        if eval_mode and not use_random_map:
            state = env.reset(data_list[eps])
        else:
            state = env.reset()
            
        if message:
            print("[Evaluation] eps:{} start!".format(eps))

        if config['use_recurrent_method']:
            hidden_out = torch.zeros([1, 1, config['recurrent_hidden_dim']], dtype=torch.float).cuda()

        start_time = time.time()
        while True:
            # Choose action and run
            if config['use_recurrent_method']:
                hidden_in = hidden_out
                action, hidden_out = rl_core.choose_action(state, hidden_in, eval=True)
            else:
                action = rl_core.choose_action(state, eval=True)
            state_next, reward, done, info = env.step(action)
            
            if not eval_mode:
                if eps < record_len:
                    im = env.render(render=render)
                    im_pil = Image.fromarray(cv2.cvtColor(np.uint8(im*255), cv2.COLOR_BGR2RGB))
                    images.append(im_pil)
            else:
                eps_num_per_map = total_eps // total_map_num
                if eps % eps_num_per_map == 0 or render:
                    im = env.render(render=render)
                    im_pil = Image.fromarray(cv2.cvtColor(np.uint8(im*255), cv2.COLOR_BGR2RGB))
                    images.append(im_pil)

            acc_reward += reward
            
            state = state_next.copy()
            step += 1

            end_time = time.time()
            if info['result'] == 'arrival':
                success_count += 1
                epi_result = 'Success'
                success_epi_list.append(eps)
                path_length.append(env.car.get_path_length())
                arrival_time.append(end_time-start_time)
                break
            elif info['result'] == 'collision':
                collision_count += 1
                epi_result = 'Collision'
                collision_epi_list.append(eps)
                path_length.append(-1)
                arrival_time.append(end_time-start_time)
                break
            elif step > config['max_epi_len']:
                timeout_count += 1
                epi_result = 'Timeout'
                timeout_epi_list.append(eps)
                path_length.append(-1)
                arrival_time.append(end_time-start_time)
                break

            if extract_video:
                if not asked:
                    k = cv2.waitKey(0)
                    if k == ord('n'):
                        print(f'skip current episode')
                        break
                    else:
                        print(f'Play')
                        asked = True

        if extract_video:
            # For recording
            print(f'Save(s) or Skip(k)? Press: \'s\' or \'k\'')
            k = cv2.waitKey(0)
            if k == ord("s"):
                print(f'Saving ...')
                gif_path = './tmp_arbit/'
                gif_file_name = gif_name[:-4] + '_' + str(record_count) + gif_name[-4:]
                images[0].save(gif_path + gif_file_name,
                    save_all=True, append_images=images[1:], optimize=True, duration=40, loop=0)
                record_count += 1
            else:
                print(f'Skip ...')
            images = []
            continue

        # log the result of the episode
        total_res_list.append(epi_result)
        
        # End episode
        avg_reward_list.append(acc_reward/step)

        if message:
            print("[Evaluation] eps:{} end! Result: {}".format(eps, epi_result))
            print("Current result: Success Rate: {}, Collision Rate: {}, Timeout Rate: {}, Path Length: {}, Arrival Time: {}." \
                .format(success_count/total_eps, collision_count/total_eps, timeout_count/total_eps, \
                        sum(path_length)/len(path_length), sum(arrival_time)/len(arrival_time)))

    if log_path is not None:
        record_result(total_res_list, total_map_num, log_path)

    if gif_path is not None:
        print("Save evaluation GIF ...")
        images[0].save(gif_path + gif_name,
            save_all=True, append_images=images[1:], optimize=True, duration=40, loop=0)

    rl_core.actor.train()

    avg_rewards = sum(avg_reward_list) / len(avg_reward_list)
    success_rate, collision_rate, timeout_rate = success_count/total_eps, collision_count/total_eps, timeout_count/total_eps
    if len(path_length) == 0:
        avg_path_length = -1
    else:
        avg_path_length = sum(path_length) / len(path_length)
    if len(arrival_time) == 0:
        avg_arrival_time = -1
    else:
        avg_arrival_time = sum(arrival_time) / len(arrival_time)

    return avg_rewards, success_rate, collision_rate, timeout_rate, avg_path_length, avg_arrival_time


def seed_everything(seed):
    torch.manual_seed(seed)
    random.seed(seed)
    np.random.seed(seed)
    print(f'Set random seed: {seed}')

def eval_all_disturbance_configs(configs):
    # for i in all_sensor_configs:
        pass

def record_result(result, map_num, log_path):
    result = [i.lower() for i in result]
    map_num_per_eps = len(result) // map_num
    res_per_map_list = []

    for i in range(0, len(result) // map_num_per_eps):
        res_per_map_list.append(result[i*map_num_per_eps:(i+1)*map_num_per_eps])

    sr_total, cr_total, tr_total = [], [], []
    with open(log_path, 'wt') as f:
        for i in range(len(result) // map_num_per_eps):
            sr = res_per_map_list[i].count('success')  #/ map_num_per_eps
            cr = res_per_map_list[i].count('collision')  #/ map_num_per_eps
            tr = res_per_map_list[i].count('timeout')  #/ map_num_per_eps

            sr_total.append(sr)
            cr_total.append(cr)
            tr_total.append(tr)

            log_text = 'Map' + str(i+1) + ' [S/C/T]: ' + str(sr) + ' / ' + str(cr) + ' / ' + str(tr) + '\n'
            f.write(log_text)

        # f.write('=total avg=\n')
        f.write('Total success: ' + str(result.count('success')) + '\n')
        f.write('Total collision: ' + str(result.count('collision')) + '\n')
        f.write('Total timeout: ' + str(result.count('timeout')) + '\n')

        all_result = ''
        for i, res in enumerate(result):
            if i % 5 == 0 and i > 0:
                all_result +='\n'
            all_result += res + ' '
        f.write(all_result)

    print(f'Log file saved.')

if __name__ == "__main__":
    # Set configs
    config = None
    with open('configs.yaml') as f:
        config = yaml.load(f, Loader=yaml.FullLoader)

    stack_N = config['stack_N']
    trained_model_path = ""
    path_num = '10000'
    
    anet_path = trained_model_path + "sac_anet_" + path_num + ".pt"
    cnet_path = trained_model_path + "sac_cnet_" + path_num + ".pt"
    epnet_path = trained_model_path + "sac_epnet_" + path_num + ".pt"

    eval_filename=""

    disturb_name = 'no_dist'

    log_path = 'exp_result/stack_N_' + str(stack_N) + '_' + disturb_name + '_disturbed.txt'
    gif_name = 'ours_stack_N_' + str(stack_N) + '_' + disturb_name + '_disturb.gif'
 
    reward_decay = config['reward_decay']
    actor_lr = config['actor_lr']
    critic_lr = config['critic_lr']
    batch_size = config['batch_size']
    memory_size = config['memory_size']

    actor_model = conv_model.PolicyNetGaussian
    if config['use_double_q_net']:
        critic_model = conv_model.DoubleQNet
    else:
        critic_model = conv_model.QNet
    
    rl_core = sac.SAC(
        model = [actor_model, critic_model],
        n_actions = 2,
        learning_rate = [actor_lr, critic_lr],
        reward_decay = reward_decay,
        memory_size = memory_size,
        batch_size = batch_size,
        alpha = 0.1,
        config = config,
        auto_entropy_tuning=True)
    if config['use_cmap']:
        rl_core.init_epnet()

    if config['use_cmap']:
        rl_core.save_load_model("load", anet_path, cnet_path, epnet_path, load_for_eval=True)
    else:
        rl_core.save_load_model("load", anet_path, cnet_path, "", load_for_eval=True)

    print("=======================================================")
    print("Evaluation start!")
    print("=======================================================")
    n_cpu = os.cpu_count()
    print(f'Currently used cpu: {n_cpu}')
    ray.init(num_cpus=n_cpu)

    avg_rewards, success_rate, collision_rate, timeout_rate, path_length, arrival_time = \
        run(rl_core, total_eps=250, message=True, render=False, \
        map_path="maps/map3_2.png",
        config=config,
        gif_name=gif_name,
        eval_filename=eval_filename,
        eval_mode=True,
        use_random_map=True,
        log_path=log_path)

    print("=======================================================")
    print("Evaluation done!")
    print("Avg rewards: {}".format(avg_rewards))
    print("Success Rate: {}".format(success_rate))
    print("Collision Rate: {}".format(collision_rate))
    print("timeout Rate: {}".format(timeout_rate))
    print("Path Length: {}".format(path_length))
    print("Arrival Time: {}".format(arrival_time))
    print("=======================================================")