import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F

from train_epnet import EPNetTrainer
from entropy_prediction_net import EPNet

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

class SAC():
    def __init__(
        self,
        model,
        n_actions,
        learning_rate = [1e-4, 2e-4],
        reward_decay = 0.99,
        memory_size = 5000,
        batch_size = 64,
        tau = 0.005,
        alpha = 0.5,
        config = None,
        auto_entropy_tuning = True,
        criterion = nn.MSELoss(),
        beta = 0.1,
        entropy_lr = 0.0003
    ):
        # initialize parameters
        self.n_actions = n_actions
        self.lr = learning_rate
        self.gamma = reward_decay
        self.memory_size = memory_size
        self.batch_size = batch_size
        self.tau = tau
        self.alpha = alpha
        self.config = config
        self.auto_entropy_tuning = auto_entropy_tuning
        self.criterion = criterion
        self.entropy_lr = entropy_lr
        self._build_net(model[0], model[1])
        self.init_memory()
        self.beta = beta

        self.dummy_zero = torch.zeros(self.config['batch_size']).to(device)

    def _build_net(self, anet, cnet):
        # Policy Network
        self.actor = anet(self.config).to(device)
        self.actor_optim = optim.Adam(self.actor.parameters(), lr=self.lr[0])

        # Evaluation Critic Network (new)
        self.critic = cnet(self.config).to(device)
        self.critic_optim = optim.Adam(self.critic.parameters(), lr=self.lr[1])

        # Target Critic Network (old)
        self.critic_target = cnet(self.config).to(device)
        for target_param, param in zip(self.critic_target.parameters(), self.critic.parameters()):
            target_param.data.copy_(param.data)
        for p in self.critic_target.parameters():
            p.requires_grad = False

        if self.auto_entropy_tuning:
            self.target_entropy = -torch.tensor(self.n_actions).to(device)
            self.log_alpha = torch.zeros(1, requires_grad=True, device=device)
            self.alpha_optim = optim.Adam([self.log_alpha], lr=self.entropy_lr)

    def save_load_model(self, op, anet_path, cnet_path, epnet_path=False, alpha_path=None, load_for_eval=False, lazy_load_epnet=False):
        if op == "save":
            torch.save(self.critic.state_dict(), cnet_path)
            torch.save(self.actor.state_dict(), anet_path)
            torch.save([self.target_entropy, self.log_alpha], alpha_path)
            if self.config['use_cmap']:
                torch.save(self.epnet.state_dict(), epnet_path)
        elif op == "load":
            self.critic.load_state_dict(torch.load(cnet_path, map_location=device))
            self.critic_target.load_state_dict(torch.load(cnet_path, map_location=device))
            self.actor.load_state_dict(torch.load(anet_path, map_location=device))

            if not load_for_eval:
                self.target_entropy, self.log_alpha = torch.load(alpha_path, map_location=device)
            if self.config['use_cmap'] and not lazy_load_epnet:
                self.epnet.load_state_dict(torch.load(epnet_path, map_location=device))

    def save_load_optim(self, op, actor_optim_path, critic_optim_path, alpha_optim_path):
        if op == "save":
            torch.save(self.critic_optim.state_dict(), critic_optim_path)
            torch.save(self.actor_optim.state_dict(), actor_optim_path)
            torch.save(self.alpha_optim.state_dict(), alpha_optim_path)
        elif op == "load":
            self.critic_optim.load_state_dict(torch.load(critic_optim_path, map_location=device))
            self.actor_optim.load_state_dict(torch.load(actor_optim_path, map_location=device))
            self.alpha_optim.load_state_dict(torch.load(alpha_optim_path, map_location=device))

    def choose_action(self, s, eval=False):
        s_ts = torch.FloatTensor(np.expand_dims(s,0)).to(device)
        if eval == False:
            
            ###           Use confidence mapping
            if self.config['use_confidence_mapping']:
                action, _, _ = self.actor.sample(s_ts, self.split_vc_from_state(s_ts))
            else:
                action, _, _ = self.actor.sample(s_ts)
        
        else:
            if self.config['use_confidence_mapping']:
                _, _, action = self.actor.sample(s_ts, self.split_vc_from_state(s_ts))
            else:
                _, _, action = self.actor.sample(s_ts)
        
        action = action.cpu().detach().numpy()[0]
        return action

    def init_memory(self):
        self.memory_counter = 0
        self.memory = {"s":[], "a":[], "r":[], "sn":[], "end":[]}

    def store_transition(self, s, a, r, sn, end):
        if self.memory_counter <= self.memory_size:
            self.memory["s"].append(s)
            self.memory["a"].append(a)
            self.memory["r"].append(r)
            self.memory["sn"].append(sn)
            self.memory["end"].append(end)
        else:
            index = self.memory_counter % self.memory_size
            self.memory["s"][index] = s
            self.memory["a"][index] = a
            self.memory["r"][index] = r
            self.memory["sn"][index] = sn
            self.memory["end"][index] = end

        self.memory_counter += 1

    def soft_update(self):
        with torch.no_grad():
            for target_param, eval_param in zip(self.critic_target.parameters(), self.critic.parameters()):
                target_param.copy_((1 - self.tau)*target_param.data + self.tau*eval_param.data)

    def init_epnet(self):
        self.epnet = EPNet(self.config).to(device)
        self.epnet_trainer = EPNetTrainer(self.epnet, device, self.config)

    def train_epnet(self):
        s_ts = np.array(self.memory["s"])
        a_ts = np.array(self.memory["a"])
        sn_ts = np.array(self.memory["sn"])

        self.epnet_trainer.main(
            self.split_vc_from_state(s_ts), 
            a_ts, 
            self.split_vel_from_state(s_ts), 
            self.split_vc_from_state(sn_ts),
            "",
            "",
            False,
            "",
            True
            )    

    def split_vc_from_state(self, s_ts):
        vc = s_ts[:, self.config['obs_dim']*self.config['stack_N']+self.config['state_dim'] \
            : self.config['obs_dim']*self.config['stack_N']+self.config['state_dim']+self.config['cmap_sector_num']]
        return vc

    def split_vel_from_state(self, s_ts):
        vel = s_ts[:, self.config['obs_dim']*self.config['stack_N']+self.config['state_dim'] - self.config['act_dim'] \
            : self.config['obs_dim']*self.config['stack_N']+self.config['state_dim']]
        return vel

    def load_epnet(self, epnet_path):
        self.epnet = EPNet(self.config).to(device)
        self.epnet.load_state_dict(torch.load(epnet_path, map_location=device)['model'])

    def learn(self):
        # sample batch memory from all memory
        if self.memory_counter > self.memory_size:
            sample_index = np.random.choice(self.memory_size, size=self.batch_size)
        else:
            sample_index = np.random.choice(self.memory_counter, size=self.batch_size)
        
        s_batch = [self.memory["s"][index] for index in sample_index]
        a_batch = [self.memory["a"][index] for index in sample_index]
        r_batch = [self.memory["r"][index] for index in sample_index]
        sn_batch = [self.memory["sn"][index] for index in sample_index]
        end_batch = [self.memory["end"][index] for index in sample_index]

        # Construct torch tensor
        s_ts = torch.FloatTensor(np.array(s_batch)).to(device)
        a_ts = torch.FloatTensor(np.array(a_batch)).to(device)
        r_ts = torch.FloatTensor(np.array(r_batch)).to(device).view(self.batch_size, 1)
        sn_ts = torch.FloatTensor(np.array(sn_batch)).to(device)
        end_ts = torch.FloatTensor(np.array(end_batch)).to(device).view(self.batch_size, 1)
        
        # TD-target
        with torch.no_grad():
            if self.config['use_confidence_mapping']:
                a_next, logpi_next, _ = self.actor.sample(sn_ts, self.split_vc_from_state(sn_ts), self.dummy_zero)
            else:
                a_next, logpi_next, _ = self.actor.sample(sn_ts)
            if self.config['use_double_q_net']:
                critic1_pi_target, critic2_pi_target = self.critic_target(sn_ts, a_next)
                critic_pi_target = torch.min(critic1_pi_target, critic2_pi_target)
            else:
                critic_pi_target = self.critic_target(sn_ts, a_next)
            critic_target = r_ts + (1 - end_ts) * self.gamma * (critic_pi_target - self.alpha * logpi_next)
        
        # Critic loss
        if self.config['use_double_q_net']:
            critic1_eval, critic2_eval = self.critic(s_ts, a_ts)
            critic1_loss = self.criterion(critic1_eval, critic_target)
            critic2_loss = self.criterion(critic2_eval, critic_target)
            self.critic_loss = critic1_loss + critic2_loss
        else:
            critic_eval = self.critic(s_ts, a_ts)
            self.critic_loss = self.criterion(critic_eval, critic_target)

        self.critic_optim.zero_grad()
        self.critic_loss.backward()
        self.critic_optim.step()

        # Actor loss
        if self.config['use_cmap']:
            self.vc_ts = self.split_vc_from_state(s_ts)
            self.vel_ts = self.split_vel_from_state(s_ts)
        if self.config['use_confidence_mapping']:
            a_curr, logpi_curr, _ = self.actor.sample(s_ts, self.vc_ts, self.dummy_zero)
        else:
            a_curr, logpi_curr, _ = self.actor.sample(s_ts)
        if self.config['use_double_q_net']:
            critic1_current, critic2_current = self.critic(s_ts, a_curr)
            critic_current = torch.min(critic1_current, critic2_current)
        else:
            critic_current = self.critic(s_ts, a_curr)

        if self.config['use_cmap']:
            H_hat = self.epnet.predict(self.vc_ts, a_ts, self.vel_ts)
            self.actor_loss = ((self.alpha*logpi_curr) - critic_current - self.beta * H_hat).mean()
        else:
            self.actor_loss = ((self.alpha*logpi_curr) - critic_current).mean()

        self.actor_optim.zero_grad()
        self.actor_loss.backward()
        self.actor_optim.step()

        self.soft_update()
        
        # Adaptive entropy adjustment
        if self.auto_entropy_tuning:
            alpha_loss = -(self.log_alpha * (logpi_curr + self.target_entropy).detach()).mean()
            self.alpha_optim.zero_grad()
            alpha_loss.backward()
            self.alpha_optim.step()
            self.alpha = float(self.log_alpha.exp().detach().cpu().numpy())
        
        return float(self.actor_loss.detach().cpu().numpy()), float(self.critic_loss.detach().cpu().numpy())