from time import time
from mobile2d.lidar_model import LidarModel
from mobile2d.wmr_model import KinematicModel
from envs.disturbed_sensor import DisturbedSensor
from envs.confidence_map import ConfidenceMap
import cv2
import numpy as np
import matplotlib.pyplot as plt
import time
import ray


def test_env(config):
    ##############################################################################################
    #                                       Initialization
    ##############################################################################################
    img = cv2.flip(cv2.imread(config['map_path']), 0)
    img[img>128] = 255
    img[img<=128] = 0
    m = np.asarray(img)
    m = cv2.cvtColor(m, cv2.COLOR_RGB2GRAY)
    m = m.astype(np.float32) / 255.
    m_origin = m.copy()
    img = img.astype(np.float32) / 255.

    # Set lidar model
    lmodel = LidarModel(m, config=config) # Set sensor model 
    car = KinematicModel(v_range=config['linear_vel'], w_range=config['angular_vel'], \
                         d=12, wu=12, wv=4, car_w=18, car_f=14, car_r=10, dt=config['dt']) # Set car model
    
    # Set car
    car.x, car.y, car.yaw = 200,100, 70

    # Set cmap
    cmap = ConfidenceMap(m, config)

    # Set disturbance
    disturbed_sensor = DisturbedSensor(config)

    # Set pyplot
    plt.ion()
    fig = plt.figure()
    # plt.ylim(-0.1,1.1)
    # fig.canvas.set_window_title('Vectorized Confidence')
    ax = fig.add_subplot(2,1,1)
    ax2 = fig.add_subplot(2,1,2)
    ax.set_ylim([-0.1, 1.5])
    ax2.set_ylim([0, 60])
    
    x = np.arange(0, config['cmap_sector_num'])
    y = np.zeros(config['cmap_sector_num'])

    x_ax2, y_ax2 = [], []

    line, = ax.plot(x, y, 'b-')
    line2, = ax2.plot(x_ax2, y_ax2, 'b-')
    
    sensor_data = np.zeros((config['stack_N']*config['obs_dim']))

    total_step = 0

    while True:
        img_ = img.copy()
        print("\rx={}, y={}, v={}, yaw={}, w={}".format(str(car.x)[:5],str(car.y)[:5],\
                str(car.v)[:5],str(car.yaw)[:5],str(car.w)[:5]), end="\t")

        # Render car
        car.update()
        pos = (car.x, car.y, car.yaw)

        # Render sensor
        m = m_origin.copy()
        lmodel.img_map = m
        sdata, plist = lmodel.measure_2d(pos)

        # trigger Disturbance
        sdata, disturb_mask = disturbed_sensor.trigger_both_disturbance(sdata)
        sdata, plist = lmodel.measure_2d(pos, sdata)

        sensor_data[config['obs_dim']:] = sensor_data[:-config['obs_dim']]
        sensor_data[:config['obs_dim']] = [i/config['sensor_range'] for i in sdata]

        for pts in plist:
            cv2.line(
                img_, 
                (int(1*pos[0]), int(1*pos[1])), 
                (int(1*pts[0]), int(1*pts[1])),
                (1.0,0.5,0.0), 1)

        # Car render
        img_ = car.render(img_)

        # Render cmap
        color = (0/255., 255/255., 204/255.)
        img_ = cv2.circle(img_, (int(round(car.x)), int(round(car.y))), config['cmap_radius'], color, 1)

        # Render sensor data
        origin_H, origin_W = img_.shape[0], img_.shape[1]
        sensor_viz_H, sensor_viz_W = origin_H // 4, origin_W
        sensor_img = sensor_data[::-1].reshape(config['stack_N'], config['obs_dim']).astype(np.float32)
        sensor_img = cv2.resize(sensor_img, (sensor_viz_W, sensor_viz_H))
        sensor_img = cv2.cvtColor(sensor_img, cv2.COLOR_GRAY2BGR)

        # Render goal
        target_pos = (210,600)
        cv2.circle(img_, (int(target_pos[0]), int(target_pos[1])), 15, (1.0,0.5,0.7), 3)  

        # Stack sensor viz on the original image
        img_ = np.concatenate((img_,sensor_img), axis=0)

        img_ = cv2.flip(img_,0)
        cv2.imshow("Test Driving", img_)


        start_time = time.time()
        sensing_confidence_sectors = cmap.step(pos, plist)
        end_time = time.time()

        print(f'cmap step time: {end_time - start_time}')
        # val = sensing_confidence_sectors.sum()
        # print(val)

        c_img = np.zeros_like(cmap.sensing_cmap)
        c_img = cv2.normalize(cmap.sensing_cmap, c_img, 255, 0, cv2.NORM_MINMAX).astype(np.uint8)

        cv2.imshow('2d local cmap', c_img)

        global_cmap = cmap.m
        global_cmap = cv2.flip(global_cmap, 0)
        cv2.imshow('global cmap', global_cmap)

        # 2d cmap
        # c_img = cmap.sensing_cmap

        entropy_value = cmap.calc_sensing_val(sensing_confidence_sectors)

        print(f'Entropy: {entropy_value}')

        x_ax2.append(total_step)
        y_ax2.append(entropy_value)
        line2, = ax2.plot(x_ax2, y_ax2, 'b-')

        line.set_ydata(sensing_confidence_sectors)
        fig.canvas.draw()
        ax2.set_xlim(left=max(0, total_step-100), right=total_step+100)
        # fig.canvas.flush_events()

        total_step += 1

        # Control
        k = cv2.waitKey(1)
        if k == ord("a"):
            car.w += 10.
        elif k == ord("d"):
            car.w -= 10.
        elif k == ord("w"):
            car.v += 10.
        elif k == ord("s"):
            car.v -= 10.
        elif k == 27:  # ESC
            print()
            break


if __name__ == '__main__':
    # Set config
    import yaml
    config = None
    with open('configs.yaml') as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
    config['map_path'] = "maps/map4.png"

    # ray setting
    import os
    n_cpu = os.cpu_count()
    print(f'Currently used cpu: {n_cpu}')
    import ray
    ray.init(num_cpus=n_cpu)
    
    # Start
    test_env(config)