import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
import yaml
import gzip
import pickle
import datetime
import time
import cv2

from confidence_prediction_net import EPNet
from cmap_dataset import CmapDataset
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter

class EPNetTrainer:
    def __init__(self, epnet, device, config):
        self.config = config
        self.device = device
        self.epnet = epnet
        self.count = 0  # for updating replay buffer path
        self.criterion = nn.MSELoss()
        self.optimizer = optim.Adam(self.epnet.parameters(), lr=0.001)
        self.scheduler = optim.lr_scheduler.MultiStepLR(self.optimizer, milestones=[100,200,300,400,500], gamma=0.5)
        # self.scheduler = optim.lr_scheduler.MultiStepLR(self.optimizer, milestones=[10,220,230,240,250,260], gamma=0.5)

    def train(self, epoch, model, train_loader, criterion, optimizer, writer):
        model.train()
        train_loss = 0
        
        for idx, (vc, a_ts, vel, vc_next) in enumerate(train_loader):
            vc, a_ts, vel, vc_next = vc.to(self.device), a_ts.to(self.device), vel.to(self.device), vc_next.to(self.device)
            optimizer.zero_grad()
            vc_pred = model(vc, a_ts, vel)
            loss = criterion(vc_next, vc_pred)
            loss.backward()
            self.optimizer.step()
            train_loss += loss.item()

            self.count += 1
            if self.count >= self.dataset.size_per_buffer:
                self.count = 0
                self.dataset.update_data_path()
                self.train_data, self.test_data = train_test_split(self.dataset, test_size=0.15, shuffle=False)
                self.train_loader = DataLoader(self.train_data, batch_size=self.config['batch_size'], shuffle=False)
                self.test_loader = DataLoader(self.test_data, batch_size=self.config['batch_size'], shuffle=False)
        
            print('\r[Training] Epoch: {:>4} | Current batch: {:>4} /{:>4} | lr: {:>6.6f} | AvgLoss: {:.4f}'. \
                    format(epoch, idx, len(train_loader), optimizer.param_groups[0]['lr'], train_loss/(idx+1)), end='')
            writer.add_scalar('Training/Avg_Train_loss', train_loss/(idx+1), global_step=epoch * len(train_loader)+idx+1)
        print()

    def test(self, epoch, model, test_loader, criterion, optimizer, writer):
        model.eval()
        test_loss = 0

        with torch.no_grad():
            for idx, (vc, a_ts, vel, vc_next) in enumerate(test_loader):
                vc, a_ts, vel, vc_next = vc.to(self.device), a_ts.to(self.device), vel.to(self.device), vc_next.to(self.device)
                vc_pred = model(vc, a_ts, vel)
                loss = criterion(vc_next, vc_pred)
                test_loss += loss.item()

                print('\r[Testing]  Epoch: {:>4} | Current batch: {:>4} /{:>4} | lr: {:>6.6f} | AvgLoss: {:.4f}'. \
                    format(epoch, idx, len(test_loader), optimizer.param_groups[0]['lr'], test_loss/(idx+1)), end='')
                writer.add_scalar('Testing/Avg_Test_loss', test_loss/(idx+1), global_step=epoch * len(test_loader)+idx+1)
        print()

    def main(self, vc_ndarray, a_ts_ndarray, vel_ndarray, vc_next_ndarray, 
            trained_model_path, exp_name, resume_training, load_model_name, train_online, data_root_path):
        start_epoch = 0
        if resume_training:
            model_checkpoint = torch.load(load_model_name)
            self.epnet.load_state_dict(model_checkpoint['model'])
            self.optimizer.load_state_dict(model_checkpoint['optimizer'])
            start_epoch = model_checkpoint['epoch']
            print('\nResume training.')
        
        # Data setting
        if not train_online:
            self.dataset = CmapDataset(self.config, vc_ndarray, a_ts_ndarray, vel_ndarray, vc_next_ndarray, train_online=train_online, data_root_path=data_root_path)
        else:
            self.dataset = CmapDataset(self.config, vc_ndarray, a_ts_ndarray, vel_ndarray, vc_next_ndarray, train_online=train_online)
        
        self.train_data, self.test_data = train_test_split(self.dataset, test_size=0.15, shuffle=False)
        self.train_loader = DataLoader(self.train_data, batch_size=self.config['batch_size'], shuffle=False)
        self.test_loader = DataLoader(self.test_data, batch_size=self.config['batch_size'], shuffle=False)

        # Tensorboard setting
        now = datetime.datetime.now()
        date_time = now.strftime('%Y-%m-%d')
        writer = SummaryWriter('trained_epnet/logs/' + date_time + '/' + exp_name)

        ############################################################################################################
        #                                                   Training & Testing 
        ############################################################################################################
        print('\n===================== [Start main] =====================')
        start_time = time.time()
        for epoch in range(start_epoch, self.config['epnet_training_epochs']+1):
            self.train(epoch, self.epnet, self.train_loader, self.criterion, self.optimizer, writer)
            self.test(epoch, self.epnet, self.test_loader, self.criterion, self.optimizer, writer)
            self.scheduler.step()

            if not train_online:
                if epoch % 50 == 0 and epoch > 0:
                    sec = time.time() - start_time
                    elapsed_time = str(datetime.timedelta(seconds=sec)).split('.')[0]
                    print("Elapsed time: {}".format(elapsed_time))

                    model_filename = trained_model_path + 'model/' + exp_name + '_' + str(epoch) + '.pth'
                    save_checkpoint(epoch, self.epnet, self.optimizer, self.scheduler, model_filename)
                    print(f'Saved model!')
            
        writer.close()

    def calc_entropy(self, vc_next):
        if torch.is_tensor(vc_next):
            vc_next = vc_next.numpy()
        vc_next_sum = vc_next.sum()
        if vc_next_sum == 0:
            return 0
        p = vc_next / vc_next_sum
        id_p = np.where(p != 0)
        return -np.sum(p[id_p]*np.log(p[id_p]))

def save_checkpoint(epoch, model, optimizer, scheduler, filename):
    state = {
        'epoch': epoch,
        'model': model.state_dict(),
        'optimizer': optimizer.state_dict(),
        'scheduler': scheduler.state_dict()
    }
    torch.save(state, filename)

def split_vc_from_state(config, s_ts):
    vc = s_ts[:, config['obs_dim']*config['stack_N']+config['state_dim'] \
        : config['obs_dim']*config['stack_N']+config['state_dim']+config['cmap_sector_num']]
    return vc

def split_vel_from_state(config, s_ts):
    vel = s_ts[:, config['obs_dim']*config['stack_N']+config['state_dim'] - config['act_dim'] \
        : config['obs_dim']*config['stack_N']+config['state_dim']]
    return vel


if __name__ == '__main__':
    config = None
    with open('configs.yaml') as f:
        config = yaml.load(f, Loader=yaml.FullLoader)

    # load_model_name = './trained_epnet/model/epnet_199.pth'
    
    data_root_path = './replay_buffer/'
    load_model_name = './trained_epnet/model/epnet_279.pth'

    buffer_path = './replay_buffer/buffer_20k.pkl'
    trained_model_path = './trained_epnet/'
    exp_name = 'scp-net_trained_with_random_map'
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    epnet = EPNet(config).to(device)
    epnet_trainer = EPNetTrainer(epnet, device, config)
    
    # Start training
    epnet_trainer.main(None, 
                        None, 
                        None, 
                        None,
                        trained_model_path,
                        exp_name,
                        resume_training=False,
                        load_model_name=load_model_name,
                        train_online=False,
                        data_root_path=data_root_path
                        )